# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the 'verdi computer' command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os
from collections import OrderedDict
from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands.cmd_computer import computer_disable, computer_enable, computer_setup
from aiida.cmdline.commands.cmd_computer import computer_show, computer_list, computer_rename, computer_delete
from aiida.cmdline.commands.cmd_computer import computer_test, computer_configure, computer_duplicate
from aiida import orm


def generate_setup_options_dict(replace_args={}, non_interactive=True):
    """
    Return a OrderedDict with the key-value pairs for the command line.

    I use an ordered dict because for changing entries it's easier
    to have keys (so, a dict) but the commands might require a specific order,
    so I use an OrderedDict.

    This should be then passed to ``generate_setup_options()``.

    :param replace_args: a dictionary with the keys to replace, if needed
    :return: an OrderedDict with the command-line options
    """
    valid_noninteractive_options = OrderedDict()

    if non_interactive:
        valid_noninteractive_options['non-interactive'] = None
    valid_noninteractive_options['label'] = 'noninteractive_computer'
    valid_noninteractive_options['hostname'] = "localhost"
    valid_noninteractive_options['description'] = "my description"
    valid_noninteractive_options['transport'] = "local"
    valid_noninteractive_options['scheduler'] = "direct"
    valid_noninteractive_options['shebang'] = "#!/bin/bash"
    valid_noninteractive_options['work-dir'] = "/scratch/{username}/aiida_run"
    valid_noninteractive_options['mpirun-command'] = "mpirun -np {tot_num_mpiprocs}"
    valid_noninteractive_options['mpiprocs-per-machine'] = "2"
    # Make them multiline to test also multiline options
    valid_noninteractive_options['prepend-text'] = "date\necho 'second line'"
    valid_noninteractive_options['append-text'] = "env\necho '444'\necho 'third line'"

    # I replace kwargs here, so that if they are known, they go at the right order
    for k in replace_args:
        valid_noninteractive_options[k] = replace_args[k]

    return valid_noninteractive_options


def generate_setup_options(ordereddict):
    """
    Given an (ordered) dict, returns a list of options

    Note that at this moment the implementation only supports long options
    (i.e. --option=value) and not short ones (-o value).
    Set a value to None to avoid the '=value' part.

    :param ordereddict: as generated by ``generate_setup_options_dict()``
    :return: a list to be passed as command-line arguments.
    """
    options = []
    for key, value in ordereddict.items():
        if value is None:
            options.append('--{}'.format(key))
        else:
            options.append('--{}={}'.format(key, value))
    return options


def generate_setup_options_interactive(ordereddict):
    """
    Given an (ordered) dict, returns a list of options

    Note that at this moment the implementation only supports long options
    (i.e. --option=value) and not short ones (-o value).
    Set a value to None to avoid the '=value' part.

    :param ordereddict: as generated by ``generate_setup_options_dict()``
    :return: a list to be passed as command-line arguments.
    """
    options = []
    for key, value in ordereddict.items():
        if value is None:
            options.append(True)
        else:
            options.append(value)
    return options


class TestVerdiComputerSetup(AiidaTestCase):
    """Tests for the 'verdi computer setup' command."""

    def setUp(self):
        self.cli_runner = CliRunner()

    def test_help(self):
        self.cli_runner.invoke(computer_setup, ['--help'], catch_exceptions=False)

    def test_reachable(self):
        import subprocess as sp
        output = sp.check_output(['verdi', 'computer', 'setup', '--help'])
        self.assertIn(b'Usage:', output)

    def test_interactive(self):
        os.environ['VISUAL'] = 'sleep 1; vim -cwq'
        os.environ['EDITOR'] = 'sleep 1; vim -cwq'
        label = 'interactive_computer'

        options_dict = generate_setup_options_dict(replace_args={'label': label}, non_interactive=False)
        # In any case, these would be managed by the visual editor
        options_dict.pop('prepend-text')
        options_dict.pop('append-text')
        user_input = "\n".join(generate_setup_options_interactive(options_dict))

        result = self.cli_runner.invoke(computer_setup, input=user_input)
        self.assertIsNone(result.exception, msg="There was an unexpected exception. Output: {}".format(result.output))

        new_computer = orm.Computer.objects.get(name=label)
        self.assertIsInstance(new_computer, orm.Computer)

        self.assertEqual(new_computer.description, options_dict['description'])
        self.assertEqual(new_computer.hostname, options_dict['hostname'])
        self.assertEqual(new_computer.get_transport_type(), options_dict['transport'])
        self.assertEqual(new_computer.get_scheduler_type(), options_dict['scheduler'])
        self.assertEqual(new_computer.get_mpirun_command(), options_dict['mpirun-command'].split())
        self.assertEqual(new_computer.get_shebang(), options_dict['shebang'])
        self.assertEqual(new_computer.get_workdir(), options_dict['work-dir'])
        self.assertEqual(new_computer.get_default_mpiprocs_per_machine(), int(options_dict['mpiprocs-per-machine']))
        # For now I'm not writing anything in them
        self.assertEqual(new_computer.get_prepend_text(), "")
        self.assertEqual(new_computer.get_append_text(), "")

    def test_mixed(self):
        os.environ['VISUAL'] = 'sleep 1; vim -cwq'
        os.environ['EDITOR'] = 'sleep 1; vim -cwq'
        label = 'mixed_computer'

        options_dict = generate_setup_options_dict(replace_args={'label': label})
        options_dict_full = options_dict.copy()

        options_dict.pop('non-interactive', 'None')

        non_interactive_options_dict = {}
        non_interactive_options_dict['prepend-text'] = options_dict.pop('prepend-text')
        non_interactive_options_dict['append-text'] = options_dict.pop('append-text')
        non_interactive_options_dict['shebang'] = options_dict.pop('shebang')
        non_interactive_options_dict['scheduler'] = options_dict.pop('scheduler')

        # In any case, these would be managed by the visual editor
        user_input = "\n".join(generate_setup_options_interactive(options_dict))
        options = generate_setup_options(non_interactive_options_dict)

        result = self.cli_runner.invoke(computer_setup, options, input=user_input)
        self.assertIsNone(result.exception, msg="There was an unexpected exception. Output: {}".format(result.output))

        new_computer = orm.Computer.objects.get(name=label)
        self.assertIsInstance(new_computer, orm.Computer)

        self.assertEqual(new_computer.description, options_dict_full['description'])
        self.assertEqual(new_computer.hostname, options_dict_full['hostname'])
        self.assertEqual(new_computer.get_transport_type(), options_dict_full['transport'])
        self.assertEqual(new_computer.get_scheduler_type(), options_dict_full['scheduler'])
        self.assertEqual(new_computer.get_mpirun_command(), options_dict_full['mpirun-command'].split())
        self.assertEqual(new_computer.get_shebang(), options_dict_full['shebang'])
        self.assertEqual(new_computer.get_workdir(), options_dict_full['work-dir'])
        self.assertEqual(new_computer.get_default_mpiprocs_per_machine(),
                         int(options_dict_full['mpiprocs-per-machine']))
        # For now I'm not writing anything in them
        self.assertEqual(new_computer.get_prepend_text(), options_dict_full['prepend-text'])
        self.assertEqual(new_computer.get_append_text(), options_dict_full['append-text'])

    def test_noninteractive(self):
        """
        Main test to check if the non-interactive command works
        """
        options_dict = generate_setup_options_dict()
        options = generate_setup_options(options_dict)

        result = self.cli_runner.invoke(computer_setup, options)

        self.assertIsNone(result.exception, result.output[-1000:])
        new_computer = orm.Computer.objects.get(name=options_dict['label'])
        self.assertIsInstance(new_computer, orm.Computer)

        self.assertEqual(new_computer.description, options_dict['description'])
        self.assertEqual(new_computer.hostname, options_dict['hostname'])
        self.assertEqual(new_computer.get_transport_type(), options_dict['transport'])
        self.assertEqual(new_computer.get_scheduler_type(), options_dict['scheduler'])
        self.assertEqual(new_computer.get_mpirun_command(), options_dict['mpirun-command'].split())
        self.assertEqual(new_computer.get_shebang(), options_dict['shebang'])
        self.assertEqual(new_computer.get_workdir(), options_dict['work-dir'])
        self.assertEqual(new_computer.get_default_mpiprocs_per_machine(), int(options_dict['mpiprocs-per-machine']))
        self.assertEqual(new_computer.get_prepend_text(), options_dict['prepend-text'])
        self.assertEqual(new_computer.get_append_text(), options_dict['append-text'])

        # Test that I cannot generate twice a computer with the same label
        result = self.cli_runner.invoke(computer_setup, options)
        self.assertIsInstance(result.exception, SystemExit)
        self.assertIn("already exists", result.output)

    def test_noninteractive_optional_default_mpiprocs(self):
        """
        Check that if is ok not to specify mpiprocs-per-machine
        """
        options_dict = generate_setup_options_dict({'label': 'computer_default_mpiprocs'})
        options_dict.pop('mpiprocs-per-machine')
        options = generate_setup_options(options_dict)
        result = self.cli_runner.invoke(computer_setup, options)

        self.assertIsNone(result.exception, result.output[-1000:])

        new_computer = orm.Computer.objects.get(name=options_dict['label'])
        self.assertIsInstance(new_computer, orm.Computer)
        self.assertIsNone(new_computer.get_default_mpiprocs_per_machine())

    def test_noninteractive_optional_default_mpiprocs_2(self):
        """
        Check that if is the specified value is zero, it means unspecified
        """
        options_dict = generate_setup_options_dict({'label': 'computer_default_mpiprocs_2'})
        options_dict['mpiprocs-per-machine'] = 0
        options = generate_setup_options(options_dict)
        result = self.cli_runner.invoke(computer_setup, options)

        self.assertIsNone(result.exception, result.output[-1000:])

        new_computer = orm.Computer.objects.get(name=options_dict['label'])
        self.assertIsInstance(new_computer, orm.Computer)
        self.assertIsNone(new_computer.get_default_mpiprocs_per_machine())

    def test_noninteractive_optional_default_mpiprocs_3(self):
        """
        Check that it fails for a negative number of mpiprocs
        """
        options_dict = generate_setup_options_dict({'label': 'computer_default_mpiprocs_3'})
        options_dict['mpiprocs-per-machine'] = -1
        options = generate_setup_options(options_dict)
        result = self.cli_runner.invoke(computer_setup, options)

        self.assertIsInstance(result.exception, SystemExit)
        self.assertIn("mpiprocs_per_machine, must be positive", result.output)

    def test_noninteractive_wrong_transport_fail(self):
        """
        Check that if fails as expected for an unknown transport
        """
        options_dict = generate_setup_options_dict(replace_args={'label': 'fail_computer'})
        options_dict['transport'] = 'unknown_transport'
        options = generate_setup_options(options_dict)
        result = self.cli_runner.invoke(computer_setup, options)
        self.assertIsInstance(result.exception, SystemExit)
        self.assertIn("'unknown_transport' is not valid", result.output)

    def test_noninteractive_wrong_scheduler_fail(self):
        """
        Check that if fails as expected for an unknown transport
        """
        options_dict = generate_setup_options_dict(replace_args={'label': 'fail_computer'})
        options_dict['scheduler'] = 'unknown_scheduler'
        options = generate_setup_options(options_dict)
        result = self.cli_runner.invoke(computer_setup, options)

        self.assertIsInstance(result.exception, SystemExit)
        self.assertIn("'unknown_scheduler' is not valid", result.output)

    def test_noninteractive_invalid_shebang_fail(self):
        """
        Check that if fails as expected for an unknown transport
        """
        options_dict = generate_setup_options_dict(replace_args={'label': 'fail_computer'})
        options_dict['shebang'] = '/bin/bash'  # Missing #! in front
        options = generate_setup_options(options_dict)
        result = self.cli_runner.invoke(computer_setup, options)

        self.assertIsInstance(result.exception, SystemExit)
        self.assertIn("The shebang line should start with", result.output)

    def test_noninteractive_invalid_mpirun_fail(self):
        """
        Check that if fails as expected for an unknown transport
        """
        options_dict = generate_setup_options_dict(replace_args={'label': 'fail_computer'})
        options_dict['mpirun-command'] = 'mpirun -np {unknown_key}'
        options = generate_setup_options(options_dict)
        result = self.cli_runner.invoke(computer_setup, options)

        self.assertIsInstance(result.exception, SystemExit)
        self.assertIn("unknown replacement field 'unknown_key'", str(result.output))


class TestVerdiComputerConfigure(AiidaTestCase):

    def setUp(self):
        """Prepare computer builder with common properties."""
        from aiida.orm.utils.builders.computer import ComputerBuilder
        self.cli_runner = CliRunner()
        self.user = orm.User.objects.get_default()
        self.comp_builder = ComputerBuilder(label='test_comp_setup')
        self.comp_builder.hostname = 'localhost'
        self.comp_builder.description = 'Test Computer'
        self.comp_builder.scheduler = 'direct'
        self.comp_builder.work_dir = '/tmp/aiida'
        self.comp_builder.prepend_text = ''
        self.comp_builder.append_text = ''
        self.comp_builder.mpiprocs_per_machine = 8
        self.comp_builder.mpirun_command = 'mpirun'
        self.comp_builder.shebang = '#!xonsh'

    def test_top_help(self):
        """Test help option of verdi computer configure."""
        result = self.cli_runner.invoke(computer_configure, ['--help'], catch_exceptions=False)
        self.assertIn('ssh', result.output)
        self.assertIn('local', result.output)

    def test_reachable(self):
        """Test reachability of top level and sub commands."""
        import subprocess as sp
        output = sp.check_output(['verdi', 'computer', 'configure', '--help'])
        sp.check_output(['verdi', 'computer', 'configure', 'local', '--help'])
        sp.check_output(['verdi', 'computer', 'configure', 'ssh', '--help'])
        sp.check_output(['verdi', 'computer', 'configure', 'show', '--help'])

    def test_local_ni_empty(self):
        """
        Test verdi computer configure local <comp>

        Test twice, with comp setup for local or ssh.

         * with computer setup for local: should succeed
         * with computer setup for ssh: should fail
        """
        self.comp_builder.label = 'test_local_ni_empty'
        self.comp_builder.transport = 'local'
        comp = self.comp_builder.new()
        comp.store()

        result = self.cli_runner.invoke(computer_configure, ['local', comp.label, '--non-interactive'],
                                        catch_exceptions=False)
        self.assertTrue(comp.is_user_configured(self.user), msg=result.output)

        self.comp_builder.label = 'test_local_ni_empty_mismatch'
        self.comp_builder.transport = 'ssh'
        comp_mismatch = self.comp_builder.new()
        comp_mismatch.store()
        result = self.cli_runner.invoke(computer_configure, ['local', comp_mismatch.label, '--non-interactive'],
                                        catch_exceptions=False)
        self.assertIsNotNone(result.exception)
        self.assertIn('ssh', result.output)
        self.assertIn('local', result.output)

    def test_local_interactive(self):
        self.comp_builder.label = 'test_local_interactive'
        self.comp_builder.transport = 'local'
        comp = self.comp_builder.new()
        comp.store()

        result = self.cli_runner.invoke(computer_configure, ['local', comp.label], input='\n', catch_exceptions=False)
        self.assertTrue(comp.is_user_configured(self.user), msg=result.output)

    def test_ssh_ni_empty(self):
        """
        Test verdi computer configure ssh <comp>

        Test twice, with comp setup for ssh or local.

         * with computer setup for ssh: should succeed
         * with computer setup for local: should fail
        """
        self.comp_builder.label = 'test_ssh_ni_empty'
        self.comp_builder.transport = 'ssh'
        comp = self.comp_builder.new()
        comp.store()

        result = self.cli_runner.invoke(computer_configure, ['ssh', comp.label, '--non-interactive'],
                                        catch_exceptions=False)
        self.assertTrue(comp.is_user_configured(self.user), msg=result.output)

        self.comp_builder.label = 'test_ssh_ni_empty_mismatch'
        self.comp_builder.transport = 'local'
        comp_mismatch = self.comp_builder.new()
        comp_mismatch.store()
        result = self.cli_runner.invoke(computer_configure, ['ssh', comp_mismatch.label, '--non-interactive'],
                                        catch_exceptions=False)
        self.assertIsNotNone(result.exception)
        self.assertIn('local', result.output)
        self.assertIn('ssh', result.output)

    def test_ssh_ni_username(self):
        """Test verdi computer configure ssh <comp> --username=<username>."""
        self.comp_builder.label = 'test_ssh_ni_username'
        self.comp_builder.transport = 'ssh'
        comp = self.comp_builder.new()
        comp.store()

        username = 'TEST'
        result = self.cli_runner.invoke(computer_configure,
                                        ['ssh', comp.label, '--non-interactive', '--username={}'.format(username)],
                                        catch_exceptions=False)
        self.assertTrue(comp.is_user_configured(self.user), msg=result.output)
        self.assertEqual(orm.AuthInfo.objects.get(
            dbcomputer_id=comp.id, aiidauser_id=self.user.id).get_auth_params()['username'],
                         username)

    def test_show(self):
        """Test verdi computer configure show <comp>."""
        self.comp_builder.label = 'test_show'
        self.comp_builder.transport = 'ssh'
        comp = self.comp_builder.new()
        comp.store()

        result = self.cli_runner.invoke(computer_configure, ['show', comp.label], catch_exceptions=False)

        result = self.cli_runner.invoke(computer_configure, ['show', comp.label, '--defaults'], catch_exceptions=False)
        self.assertIn('* username', result.output)

        result = self.cli_runner.invoke(computer_configure, ['show', comp.label, '--defaults', '--as-option-string'],
                                        catch_exceptions=False)
        self.assertIn('--username=', result.output)

        config_cmd = ['ssh', comp.label, '--non-interactive']
        config_cmd.extend(result.output.replace("'", '').split(' '))
        result_config = self.cli_runner.invoke(computer_configure, config_cmd, catch_exceptions=False)
        self.assertTrue(comp.is_user_configured(self.user), msg=result_config.output)

        result_cur = self.cli_runner.invoke(computer_configure, ['show', comp.label, '--as-option-string'],
                                            catch_exceptions=False)
        self.assertIn('--username=', result.output)
        self.assertEqual(result_cur.output, result.output)


class TestVerdiComputerCommands(AiidaTestCase):
    """Testing verdi computer commands.

    Testing everything besides `computer setup`.
    """

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """Create a new computer> I create a new one because I want to configure it and I don't want to
        interfere with other tests"""
        super(TestVerdiComputerCommands, cls).setUpClass(*args, **kwargs)
        cls.computer_name = "comp_cli_test_computer"
        cls.comp = orm.Computer(
            name=cls.computer_name,
            hostname='localhost',
            transport_type='local',
            scheduler_type='direct',
            workdir='/tmp/aiida')
        cls.comp.set_default_mpiprocs_per_machine(1)
        cls.comp.store()

    def setUp(self):
        """
        Prepare the computer and user
        """
        self.user = orm.User.objects.get_default()

        # I need to configure the computer here; being 'local',
        # there should not be any options asked here
        self.comp.configure()

        assert self.comp.is_user_configured(self.user), "There was a problem configuring the test computer"
        self.cli_runner = CliRunner()

    def test_computer_test(self):
        """
        Test if the 'verdi computer test' command works

        It should work as it is a local connection
        """
        # Testing the wrong computer will fail
        result = self.cli_runner.invoke(computer_test, ['non-existent-computer'])
        # An exception should arise
        self.assertIsNotNone(result.exception)

        # Testing the right computer should pass locally
        result = self.cli_runner.invoke(computer_test, ['comp_cli_test_computer'])
        # No exceptions should arise
        self.assertIsNone(result.exception, result.output)

    def test_computer_list(self):
        """
        Test if 'verdi computer list' command works
        """
        # Check the vanilla command works
        result = self.cli_runner.invoke(computer_list, [])
        # No exceptions should arise
        self.assertIsNone(result.exception, result.output)
        # Something should be printed to stdout
        self.assertIsNotNone(result.output)

        # Check all options run
        for opt in ['-r', '--raw', '-a', '--all']:
            result = self.cli_runner.invoke(computer_list, [opt])
            # No exceptions should arise
            self.assertIsNone(result.exception, result.output)
            # Something should be printed to stdout
            self.assertIsNotNone(result.output)

    def test_computer_show(self):
        """
        Test if 'verdi computer show' command works
        """
        import traceback

        # See if we can display info about the test computer.
        result = self.cli_runner.invoke(computer_show, ['comp_cli_test_computer'])

        # No exceptions should arise
        self.assertClickResultNoException(result)
        # Something should be printed to stdout
        self.assertIsNotNone(result.output)

        # See if a non-existent computer will raise an error.
        result = self.cli_runner.invoke(computer_show, 'non_existent_computer_name')
        # Exceptions should arise
        self.assertIsNotNone(result.exception)

    def test_computer_rename(self):
        """
        Test if 'verdi computer rename' command works
        """
        from aiida.common.exceptions import NotExistent

        # See if the command complains about not getting an invalid computer
        options = ['not_existent_computer_name']
        result = self.cli_runner.invoke(computer_rename, options)
        # Exception should be raised
        self.assertIsNotNone(result.exception)

        # See if the command complains about not getting both names
        options = ['comp_cli_test_computer']
        result = self.cli_runner.invoke(computer_rename, options)
        # Exception should be raised
        self.assertIsNotNone(result.exception)

        # The new name must be different to the old one
        options = ['comp_cli_test_computer', 'comp_cli_test_computer']
        result = self.cli_runner.invoke(computer_rename, options)
        # Exception should be raised
        self.assertIsNotNone(result.exception)

        # Change a computer name successully.
        options = ['comp_cli_test_computer', 'renamed_test_computer']
        result = self.cli_runner.invoke(computer_rename, options)
        # Exception should be not be raised
        self.assertIsNone(result.exception, result.output)

        # Check that the name really was changed
        # The old name should not be available
        with self.assertRaises(NotExistent):
            orm.Computer.objects.get(name='comp_cli_test_computer')
        # The new name should be avilable
        orm.Computer.objects.get(name='renamed_test_computer')

        # Now change the name back
        options = ['renamed_test_computer', 'comp_cli_test_computer']
        result = self.cli_runner.invoke(computer_rename, options)
        # Exception should be not be raised
        self.assertIsNone(result.exception, result.output)

        # Check that the name really was changed
        # The old name should not be available
        with self.assertRaises(NotExistent):
            orm.Computer.objects.get(name='renamed_test_computer')
        # The new name should be avilable
        orm.Computer.objects.get(name='comp_cli_test_computer')

    def test_computer_delete(self):
        """
        Test if 'verdi computer delete' command works
        """
        from aiida.common.exceptions import NotExistent

        # Setup a computer to delete during the test
        comp = orm.Computer(name='computer_for_test_delete', hostname='localhost',
                            transport_type='local', scheduler_type='direct', workdir='/tmp/aiida').store()

        # See if the command complains about not getting an invalid computer
        options = ['non_existent_computer_name']
        result = self.cli_runner.invoke(computer_delete, options)
        # Exception should be raised
        self.assertIsNotNone(result.exception)

        # Delete a computer name successully.
        options = ['computer_for_test_delete']
        result = self.cli_runner.invoke(computer_delete, options)
        # Exception should be not be raised
        self.assertClickResultNoException(result)
        # Check that the computer really was deleted
        with self.assertRaises(NotExistent):
            orm.Computer.objects.get(name='computer_for_test_delete')

    def test_computer_duplicate_interactive(self):
        os.environ['VISUAL'] = 'sleep 1; vim -cwq'
        os.environ['EDITOR'] = 'sleep 1; vim -cwq'
        label = 'computer_duplicate_interactive'
        user_input = label + '\n\n\n\n\n\n\n\n\nN'
        result = self.cli_runner.invoke(computer_duplicate, [str(self.comp.pk)], input=user_input,
                                        catch_exceptions=False)
        self.assertIsNone(result.exception, result.output)

        new_computer = orm.Computer.objects.get(name=label)
        self.assertEqual(self.comp.description, new_computer.description)
        self.assertEqual(self.comp.get_hostname(), new_computer.get_hostname())
        self.assertEqual(self.comp.get_transport_type(), new_computer.get_transport_type())
        self.assertEqual(self.comp.get_scheduler_type(), new_computer.get_scheduler_type())
        self.assertEqual(self.comp.get_shebang(), new_computer.get_shebang())
        self.assertEqual(self.comp.get_workdir(), new_computer.get_workdir())
        self.assertEqual(self.comp.get_mpirun_command(), new_computer.get_mpirun_command())
        self.assertEqual(self.comp.get_default_mpiprocs_per_machine(), new_computer.get_default_mpiprocs_per_machine())
        self.assertEqual(self.comp.get_prepend_text(), new_computer.get_prepend_text())
        self.assertEqual(self.comp.get_append_text(), new_computer.get_append_text())

    def test_computer_duplicate_non_interactive(self):
        label = 'computer_duplicate_noninteractive'
        result = self.cli_runner.invoke(computer_duplicate,
                                        ['--non-interactive', '--label=' + label, str(self.comp.pk)])
        self.assertIsNone(result.exception, result.output)

        new_computer = orm.Computer.objects.get(name=label)
        self.assertEqual(self.comp.description, new_computer.description)
        self.assertEqual(self.comp.get_hostname(), new_computer.get_hostname())
        self.assertEqual(self.comp.get_transport_type(), new_computer.get_transport_type())
        self.assertEqual(self.comp.get_scheduler_type(), new_computer.get_scheduler_type())
        self.assertEqual(self.comp.get_shebang(), new_computer.get_shebang())
        self.assertEqual(self.comp.get_workdir(), new_computer.get_workdir())
        self.assertEqual(self.comp.get_mpirun_command(), new_computer.get_mpirun_command())
        self.assertEqual(self.comp.get_default_mpiprocs_per_machine(), new_computer.get_default_mpiprocs_per_machine())
        self.assertEqual(self.comp.get_prepend_text(), new_computer.get_prepend_text())
        self.assertEqual(self.comp.get_append_text(), new_computer.get_append_text())



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Drop the DbCalcState table

Revision ID: 162b99bca4a2
Revises: a603da2cc809
Create Date: 2018-11-14 08:37:13.719646

"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from alembic import op
from sqlalchemy.dialects import postgresql
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '162b99bca4a2'
down_revision = 'a603da2cc809'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table('db_dbcalcstate')


def downgrade():
    op.create_table(
        'db_dbcalcstate', sa.Column('id', sa.INTEGER(), nullable=False),
        sa.Column('dbnode_id', sa.INTEGER(), autoincrement=False, nullable=True),
        sa.Column('state', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
        sa.Column('time', postgresql.TIMESTAMP(timezone=True), autoincrement=False, nullable=True),
        sa.ForeignKeyConstraint(
            ['dbnode_id'], [u'db_dbnode.id'],
            name=u'db_dbcalcstate_dbnode_id_fkey',
            ondelete=u'CASCADE',
            initially=u'DEFERRED',
            deferrable=True), sa.PrimaryKeyConstraint('id', name=u'db_dbcalcstate_pkey'),
        sa.UniqueConstraint('dbnode_id', 'state', name=u'db_dbcalcstate_dbnode_id_state_key'))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Click parameter types for paths."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os
import click

URL_TIMEOUT_SECONDS = 10


class AbsolutePathParamType(click.Path):
    """
    The ParamType for identifying absolute Paths (derived from click.Path).
    """

    name = 'AbsolutePath'

    def convert(self, value, param, ctx):
        value = os.path.expanduser(value)
        newval = super(AbsolutePathParamType, self).convert(value, param, ctx)
        if not os.path.isabs(newval):
            raise click.BadParameter('path must be absolute')
        return newval

    def __repr__(self):
        return 'ABSOLUTEPATH'


class ImportPath(click.Path):
    """AiiDA extension of Click's Path-type to include URLs
    An ImportPath can either be a `click.Path`-type or a URL.
    :param timeout_seconds: Timeout time in seconds that a URL response is expected.
    :value timeout_seconds: Must be an int in the range [0;60], extrema included.
    If an int outside the range [0;60] is given, the value will be set to the respective extremum value.
    If any other type than int is given a TypeError will be raised.
    """

    # pylint: disable=protected-access

    def __init__(self, timeout_seconds=URL_TIMEOUT_SECONDS, **kwargs):
        super(ImportPath, self).__init__(**kwargs)

        self.timeout_seconds = timeout_seconds

    def convert(self, value, param, ctx):
        """Overwrite `convert`
        Check first if `click.Path`-type, then check if URL.
        """
        try:
            # Check if `click.Path`-type
            return super(ImportPath, self).convert(value, param, ctx)
        except click.exceptions.BadParameter:
            # Check if URL
            return self.checks_url(value, param, ctx)

    def checks_url(self, value, param, ctx):
        """Do checks for possible URL path"""
        from six.moves import urllib
        from socket import timeout

        url = value

        try:
            urllib.request.urlopen(url, data=None, timeout=self.timeout_seconds)
        except (urllib.error.URLError, urllib.error.HTTPError, timeout):
            self.fail(
                '{0} "{1}" could not be reached within {2} s.\n'
                'It may be neither a valid {3} nor a valid URL.'.format(
                    self.path_type, click._compat.filename_to_ui(url), self.timeout_seconds, self.name), param, ctx)

        return url

    @property
    def timeout_seconds(self):
        return self._timeout_seconds

    # pylint: disable=attribute-defined-outside-init
    @timeout_seconds.setter
    def timeout_seconds(self, value):
        try:
            self._timeout_seconds = int(value)
        except ValueError:
            raise TypeError("timeout_seconds should be an integer but got: {}".format(type(value)))

        if self._timeout_seconds < 0:
            self._timeout_seconds = 0

        if self._timeout_seconds > 60:
            self._timeout_seconds = 60



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Module to define custom click param type for multiple values
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click


class MultipleValueParamType(click.ParamType):
    """
    An extension of click.ParamType that can parse multiple values for a given ParamType
    """

    def __init__(self, param_type):
        super(MultipleValueParamType, self).__init__()
        self._param_type = param_type

        if hasattr(param_type, 'name'):
            self.name = '{}...'.format(param_type.name)
        else:
            self.name = '{}...'.format(param_type.__name__.upper())

    def get_metavar(self, param):
        try:
            return self._param_type.get_metavar(param)
        except AttributeError:
            return self.name

    def convert(self, value, param, ctx):
        try:
            return tuple([self._param_type(entry) for entry in value])
        except ValueError:
            self.fail('could not convert {} into type {}'.format(value, self._param_type))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for all common top level AiiDA entity classes and methods"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import typing

from plumpy.base.utils import super_check, call_with_super_check

from aiida.common import exceptions
from aiida.common import datastructures
from aiida.common.lang import classproperty, type_check
from aiida.manage.manager import get_manager

__all__ = ('Entity', 'Collection')

EntityType = typing.TypeVar('EntityType')  # pylint: disable=invalid-name


class Collection(typing.Generic[EntityType]):
    """Container class that represents the collection of objects of a particular type."""

    # A store for any backend specific collections that already exist
    _COLLECTIONS = datastructures.LazyStore()

    @classmethod
    def get_collection(cls, entity_type, backend):
        """
        Get the collection for a given entity type and backend instance

        :param entity_type: the entity type e.g. User, Computer, etc
        :param backend: the backend instance to get the collection for
        :return: the collection instance
        """
        # Lazily get the collection i.e. create only if we haven't done so yet
        return cls._COLLECTIONS.get((entity_type, backend), lambda: entity_type.Collection(backend, entity_type))

    def __init__(self, backend, entity_class):
        """Construct a new entity collection"""
        assert issubclass(entity_class, Entity), "Must provide an entity type"
        self._backend = backend or get_manager().get_backend()
        self._entity_type = entity_class

    def __call__(self, backend):
        """
        Create a new objects collection using a different backend

        :param backend: the backend to use
        :return: a new collection with the different backend
        """
        if backend is self._backend:
            # Special case if they actually want the same collection
            return self

        return self.get_collection(self.entity_type, backend)

    @property
    def backend(self):
        """Return the backend."""
        return self._backend

    @property
    def entity_type(self):
        return self._entity_type

    def query(self):
        """
        Get a query builder for the objects of this collection

        :return: a new query builder instance
        :rtype: :class:`aiida.orm.QueryBuilder`
        """
        # pylint: disable=no-self-use
        from . import querybuilder

        query = querybuilder.QueryBuilder()
        query.append(self._entity_type, project='*')
        return query

    def get(self, **filters):
        """
        Get a single collection entry that matches the filter criteria

        :param filters: the filters identifying the object to get
        :return: the entry
        """
        res = self.find(filters=filters)
        if not res:
            raise exceptions.NotExistent("No {} with filter '{}' found".format(self.entity_type.__name__, filters))
        if len(res) > 1:
            raise exceptions.MultipleObjectsError("Multiple {}s found with the same id '{}'".format(
                self.entity_type.__name__, id))

        return res[0]

    def find(self, filters=None, order_by=None, limit=None):
        """
        Find collection entries matching the filter criteria

        :param filters: the keyword value pair filters to match
        :param order_by: a list of (key, direction) pairs specifying the sort order
        :type order_by: list
        :param limit: the maximum number of results to return
        :type limit: int
        :return: a list of resulting matches
        :rtype: list
        """
        query = self.query()
        filters = filters or {}
        query.add_filter(self.entity_type, filters)
        if order_by:
            query.order_by({self.entity_type: order_by})
        if limit:
            query.limit(limit)

        return [_[0] for _ in query.all()]

    def all(self):
        """
        Get all entities in this collection

        :return: A collection of users matching the criteria
        """
        return [_[0] for _ in self.query().all()]


class Entity(object):  # pylint: disable=useless-object-inheritance
    """An AiiDA entity"""

    _objects = None

    # Define our collection type
    Collection = Collection

    @classproperty
    def objects(cls, backend=None):  # pylint: disable=no-self-use, no-self-argument
        """
        Get an collection for objects of this type.

        :param backend: the optional backend to use (otherwise use default)
        :return: an object that can be used to access entities of this type
        """
        backend = backend or get_manager().get_backend()
        return cls.Collection.get_collection(cls, backend)

    @classmethod
    def get(cls, **kwargs):
        # pylint: disable=redefined-builtin, invalid-name
        return cls.objects.get(**kwargs)  # pylint: disable=no-member

    @classmethod
    def from_backend_entity(cls, backend_entity):
        """
        Construct an entity from a backend entity instance

        :param backend_entity: the backend entity
        :return: an AiiDA entity instance
        """
        from . import implementation

        type_check(backend_entity, implementation.BackendEntity)
        entity = cls.__new__(cls)
        entity.init_from_backend(backend_entity)
        call_with_super_check(entity.initialize)
        return entity

    def __init__(self, backend_entity):
        """
        :param backend_entity: the backend model supporting this entity
        :type backend_entity: :class:`aiida.orm.implementation.BackendEntity`
        """
        self._backend_entity = backend_entity
        call_with_super_check(self.initialize)

    def init_from_backend(self, backend_entity):
        """
        :param backend_entity: the backend model supporting this entity
        :type backend_entity: :class:`aiida.orm.implementation.BackendEntity`
        """
        self._backend_entity = backend_entity

    @super_check
    def initialize(self):
        """Initialize instance attributes.

        This will be called after the constructor is called or an entity is created from an existing backend entity.
        """

    @property
    def id(self):
        """Return the id for this entity.

        This identifier is guaranteed to be unique amongst entities of the same type for a single backend instance.

        :return: the entity's id
        """
        # pylint: disable=redefined-builtin, invalid-name
        return self._backend_entity.id

    @property
    def pk(self):
        """Return the primary key for this entity.

        This identifier is guaranteed to be unique amongst entities of the same type for a single backend instance.

        :return: the entity's principal key
        """
        return self.id

    @property
    def uuid(self):
        """Return the UUID for this entity.

        This identifier is unique across all entities types and backend instances.

        :return: the entity uuid
        :rtype: :class:`uuid.UUID`
        """
        return self._backend_entity.uuid

    def store(self):
        """Store the entity."""
        self._backend_entity.store()
        return self

    @property
    def is_stored(self):
        """Return whether the entity is stored.

        :return: boolean, True if stored, False otherwise
        :rtype: bool
        """
        return self._backend_entity.is_stored

    @property
    def backend(self):
        """
        Get the backend for this entity
        :return: the backend instance
        """
        return self._backend_entity.backend

    @property
    def backend_entity(self):
        """
        Get the implementing class for this object

        :return: the class model
        """
        return self._backend_entity



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi data remote` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import io
import click

from aiida.cmdline.commands.cmd_data import verdi_data
from aiida.cmdline.params import arguments, types
from aiida.cmdline.utils import echo
from aiida.common.files import get_mode_string


@verdi_data.group('remote')
def remote():
    """Managing RemoteData objects."""


@remote.command('ls')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:remote',)))
@click.option('-l', '--long', 'ls_long', is_flag=True, default=False, help='Display also file metadata.')
@click.option('-p', '--path', type=click.STRING, default='.', help='The folder to list.')
def remote_ls(ls_long, path, datum):
    """List directory content on remote RemoteData objects."""
    import datetime
    try:
        content = datum.listdir_withattributes(path=path)
    except (IOError, OSError) as err:
        echo.echo_critical('Unable to access the remote folde or file, check if it exists.\n'
                           'Original error: {}'.format(str(err)))
    for metadata in content:
        if ls_long:
            mtime = datetime.datetime.fromtimestamp(metadata['attributes'].st_mtime)
            pre_line = '{} {:10}  {}  '.format(
                get_mode_string(metadata['attributes'].st_mode), metadata['attributes'].st_size,
                mtime.strftime("%d %b %Y %H:%M"))
            click.echo(pre_line, nl=False)
        if metadata['isdir']:
            click.echo(click.style(metadata['name'], fg='blue'))
        else:
            click.echo(metadata['name'])


@remote.command('cat')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:remote',)))
@click.argument('path', type=click.STRING)
def remote_cat(datum, path):
    """Show the content of remote files in RemoteData objects."""
    import os
    import sys
    import tempfile
    try:
        with tempfile.NamedTemporaryFile(delete=False) as tmpf:
            tmpf.close()
            datum.getfile(path, tmpf.name)
            with io.open(tmpf.name, encoding='utf8') as fhandle:
                sys.stdout.write(fhandle.read())
    except IOError as err:
        echo.echo_critical('{}: {}'.format(err.errno, str(err)))

    try:
        os.remove(tmpf.name)
    except OSError:
        # If you cannot delete, ignore (maybe I didn't manage to create it in the first place
        pass


@remote.command('show')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:remote',)))
def remote_show(datum):
    """Show information on a RemoteData object."""
    click.echo('- Remote computer name:')
    click.echo('  {}'.format(datum.get_computer_name()))
    click.echo('- Remote folder full path:')
    click.echo('  {}'.format(datum.get_remote_path()))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Renaming `DbNode.type` to `DbNode.node_type`

Revision ID: 5ddd24e52864
Revises: d254fdfed416
Create Date: 2019-02-22 17:09:57.715114

"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# Remove when https://github.com/PyCQA/pylint/issues/1931 is fixed
# pylint: disable=no-name-in-module,import-error
from alembic import op

# revision identifiers, used by Alembic.
revision = '5ddd24e52864'
down_revision = 'd254fdfed416'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.alter_column('db_dbnode', 'type', new_column_name='node_type')  # pylint: disable=no-member
    op.create_index(op.f('ix_db_dbnode_node_type'), 'db_dbnode', ['node_type'], unique=False)
    op.drop_index('ix_db_dbnode_type', table_name='db_dbnode')


def downgrade():
    """Migrations for the downgrade."""
    op.alter_column('db_dbnode', 'node_type', new_column_name='type')  # pylint: disable=no-member
    op.create_index('ix_db_dbnode_type', 'db_dbnode', ['type'], unique=False)
    op.drop_index(op.f('ix_db_dbnode_node_type'), table_name='db_dbnode')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the 'verdi code' command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os
import subprocess as sp
import traceback
from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands.cmd_code import (setup_code, delete, hide, reveal, relabel, code_list, show, code_duplicate)
from aiida.common.exceptions import NotExistent
from aiida import orm


# pylint: disable=missing-docstring
class TestVerdiCodeSetup(AiidaTestCase):
    """Tests for the 'verdi code setup' command."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestVerdiCodeSetup, cls).setUpClass(*args, **kwargs)
        orm.Computer(
            name='comp',
            hostname='localhost',
            transport_type='local',
            scheduler_type='direct',
            workdir='/tmp/aiida').store()

    def setUp(self):
        self.comp = orm.Computer.objects.get(name='comp')

        self.cli_runner = CliRunner()
        self.this_folder = os.path.dirname(__file__)
        self.this_file = os.path.basename(__file__)

    def test_help(self):
        self.cli_runner.invoke(setup_code, ['--help'], catch_exceptions=False)

    def test_reachable(self):
        output = sp.check_output(['verdi', 'code', 'setup', '--help'])
        self.assertIn(b'Usage:', output)

    def test_interactive_remote(self):
        from aiida.orm import Code
        os.environ['VISUAL'] = 'sleep 1; vim -cwq'
        os.environ['EDITOR'] = 'sleep 1; vim -cwq'
        label = 'interactive_remote'
        user_input = '\n'.join(
            [label, 'description', 'arithmetic.add', 'yes', self.comp.name, '/remote/abs/path'])
        result = self.cli_runner.invoke(setup_code, input=user_input)
        self.assertClickResultNoException(result)
        self.assertIsInstance(Code.get_from_string('{}@{}'.format(label, self.comp.name)), Code)

    def test_interactive_upload(self):
        from aiida.orm import Code
        os.environ['VISUAL'] = 'sleep 1; vim -cwq'
        os.environ['EDITOR'] = 'sleep 1; vim -cwq'
        label = 'interactive_upload'
        user_input = '\n'.join(
            [label, 'description', 'arithmetic.add', 'no', self.this_folder, self.this_file])
        result = self.cli_runner.invoke(setup_code, input=user_input)
        self.assertIsNone(result.exception, result.output)
        self.assertIsInstance(Code.get_from_string('{}'.format(label)), Code)

    def test_noninteractive_remote(self):
        from aiida.orm import Code
        label = 'noninteractive_remote'
        options = [
            '--non-interactive', '--label={}'.format(label), '--description=description',
            '--input-plugin=arithmetic.add', '--on-computer', '--computer={}'.format(self.comp.name),
            '--remote-abs-path=/remote/abs/path'
        ]
        result = self.cli_runner.invoke(setup_code, options)
        self.assertClickResultNoException(result)
        self.assertIsInstance(Code.get_from_string('{}@{}'.format(label, self.comp.name)), Code)

    def test_noninteractive_upload(self):
        from aiida.orm import Code
        label = 'noninteractive_upload'
        options = [
            '--non-interactive', '--label={}'.format(label), '--description=description',
            '--input-plugin=arithmetic.add', '--store-in-db', '--code-folder={}'.format(self.this_folder),
            '--code-rel-path={}'.format(self.this_file)
        ]
        result = self.cli_runner.invoke(setup_code, options)
        self.assertClickResultNoException(result)
        self.assertIsInstance(Code.get_from_string('{}'.format(label)), Code)

    def test_mixed(self):
        from aiida.orm import Code
        label = 'mixed_remote'
        options = ['--description=description', '--on-computer', '--remote-abs-path=/remote/abs/path']
        user_input = '\n'.join([label, 'arithmetic.add', self.comp.name])
        result = self.cli_runner.invoke(setup_code, options, input=user_input)
        self.assertClickResultNoException(result)
        self.assertIsInstance(Code.get_from_string('{}@{}'.format(label, self.comp.name)), Code)


class TestVerdiCodeCommands(AiidaTestCase):
    """Testing verdi code commands.

    Testing everything besides `code setup`.
    """

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        from aiida import orm

        super(TestVerdiCodeCommands, cls).setUpClass(*args, **kwargs)
        orm.Computer(
            name='comp',
            hostname='localhost',
            transport_type='local',
            scheduler_type='direct',
            workdir='/tmp/aiida').store()

    def setUp(self):
        from aiida import orm
        self.comp = orm.Computer.objects.get(name='comp')

        try:
            code = orm.Code.get_from_string('code')
        except NotExistent:
            code = orm.Code(
                input_plugin_name='arithmetic.add',
                remote_computer_exec=[self.comp, '/remote/abs/path'],
            )
            code.label = 'code'
            code.description = 'desc'
            code.store()
        self.code = code

        self.cli_runner = CliRunner()

    def test_hide_one(self):
        result = self.cli_runner.invoke(hide, [str(self.code.pk)])
        self.assertIsNone(result.exception, result.output)

        self.assertTrue(self.code.hidden)

    def test_reveal_one(self):
        result = self.cli_runner.invoke(reveal, [str(self.code.pk)])
        self.assertIsNone(result.exception, result.output)

        self.assertFalse(self.code.hidden)

    def test_relabel_code(self):
        result = self.cli_runner.invoke(relabel, [str(self.code.pk), 'new_code'])
        self.assertIsNone(result.exception, result.output)
        from aiida.orm import load_node
        new_code = load_node(self.code.pk)
        self.assertEqual(new_code.label, 'new_code')

    def test_relabel_code_full(self):
        self.cli_runner.invoke(relabel, [str(self.code.pk), 'new_code@comp'])
        from aiida.orm import load_node
        new_code = load_node(self.code.pk)
        self.assertEqual(new_code.label, 'new_code')

    def test_relabel_code_full_bad(self):
        result = self.cli_runner.invoke(relabel, [str(self.code.pk), 'new_code@otherstuff'])
        self.assertIsNotNone(result.exception)

    def test_delete_one(self):
        result = self.cli_runner.invoke(delete, [str(self.code.pk)])
        self.assertIsNone(result.exception, result.output)

        with self.assertRaises(NotExistent):
            from aiida.orm import Code
            Code.get_from_string('code')

    def test_code_list(self):
        # set up second code 'code2'
        from aiida.orm import Code
        try:
            code = Code.get_from_string('code2')
        except NotExistent:
            code = Code(
                input_plugin_name='templatereplacer',
                remote_computer_exec=[self.comp, '/remote/abs/path'],
            )
            code.label = 'code2'
            code.store()

        options = [
            '-A', '-a', '-o', '--input-plugin=arithmetic.add', '--computer={}'.format(self.comp.name)
        ]
        result = self.cli_runner.invoke(code_list, options)
        self.assertIsNone(result.exception, result.output)
        self.assertTrue(str(self.code.pk) in result.output, 'PK of first code should be included')
        self.assertTrue('code2' not in result.output, 'label of second code should not be included')
        self.assertTrue('comp' in result.output, 'computer name should be included')
        self.assertNotIn(result.output, "# No codes found matching the specified criteria.")

    def test_code_list_hide(self):
        self.code.hide()
        options = ['-A']
        result = self.cli_runner.invoke(code_list, options)
        self.assertIsNone(result.exception, result.output)
        self.assertTrue(self.code.full_label not in result.output, 'code should be hidden')

        options = ['-a']
        result = self.cli_runner.invoke(code_list, options)
        self.assertIsNone(result.exception, result.output)
        self.assertTrue(self.code.full_label in result.output, 'code should be shown')

    def test_code_show(self):
        result = self.cli_runner.invoke(show, [str(self.code.pk)])
        self.assertIsNone(result.exception, result.output)
        self.assertTrue(str(self.code.pk) in result.output)

    def test_code_duplicate_interactive(self):
        os.environ['VISUAL'] = 'sleep 1; vim -cwq'
        os.environ['EDITOR'] = 'sleep 1; vim -cwq'
        label = 'code_duplicate_interactive'
        user_input = label + '\n\n\n\n\n\n'
        result = self.cli_runner.invoke(code_duplicate, [str(self.code.pk)], input=user_input, catch_exceptions=False)
        self.assertIsNone(result.exception, result.output)

        from aiida.orm import Code
        new_code = Code.get_from_string(label)
        self.assertEqual(self.code.description, new_code.description)
        self.assertEqual(self.code.get_prepend_text(), new_code.get_prepend_text())
        self.assertEqual(self.code.get_append_text(), new_code.get_append_text())

    def test_code_duplicate_non_interactive(self):
        label = 'code_duplicate_noninteractive'
        result = self.cli_runner.invoke(code_duplicate, ['--non-interactive', '--label=' + label, str(self.code.pk)])
        self.assertIsNone(result.exception, result.output)

        from aiida.orm import Code
        new_code = Code.get_from_string(label)
        self.assertEqual(self.code.description, new_code.description)
        self.assertEqual(self.code.get_prepend_text(), new_code.get_prepend_text())
        self.assertEqual(self.code.get_append_text(), new_code.get_append_text())
        self.assertEqual(self.code.get_input_plugin_name(), new_code.get_input_plugin_name())


class TestVerdiCodeCommands(AiidaTestCase):
    def setUp(self):
        self.cli_runner = CliRunner()
    def test_code_list_with_no_codes_one_error_message(self):
        result = self.cli_runner.invoke(code_list)
        self.assertEqual(1, result.output.count("# No codes found matching the specified criteria."))


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

from django.db import migrations
from aiida.backends.djsite.db.migrations import upgrade_schema_version


REVISION = '1.0.6'
DOWN_REVISION = '1.0.5'


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0005_add_cmtime_indices'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='dbpath',
            name='child',
        ),
        migrations.RemoveField(
            model_name='dbpath',
            name='parent',
        ),
        migrations.RemoveField(
            model_name='dbnode',
            name='children',
        ),
        migrations.DeleteModel(
            name='DbPath',
        ),
        migrations.RunSQL("""
            DROP TRIGGER IF EXISTS autoupdate_tc ON db_dblink;
            DROP FUNCTION IF EXISTS update_tc();
        """),
        upgrade_schema_version(REVISION, DOWN_REVISION)

    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Unit tests for the ConditionalOption."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import unittest

import click
from click.testing import CliRunner

from .conditional import ConditionalOption


class ConditionalOptionTest(unittest.TestCase):
    """Unit tests for ConditionalOption."""

    @classmethod
    def setUpClass(cls):
        cls.runner = CliRunner()

    def simple_cmd(self, pname, required_fn=lambda ctx: ctx.params.get('on'), **kwargs):
        """
        returns a command with two options:

            * an option created from the args and kwargs
            * --opt, ConditionalOption with required_fn from kwargs
        """

        # pylint: disable=no-self-use

        @click.command()
        @click.option(pname, **kwargs)
        @click.option('--opt', required_fn=required_fn, cls=ConditionalOption)
        def cmd(on, opt):
            """dummy command for testing"""
            # pylint: disable=unused-argument, invalid-name

            click.echo(opt)

        return cmd

    def test_switch_off(self):
        """
        scenario: switch --on/--off detrmines if option opt is required
        action: invoke with no options
        behaviour: flag is off by default -> command runs without complaining
        """

        cmd = self.simple_cmd('--on/--off')
        runner = CliRunner()
        result = runner.invoke(cmd, [])
        self.assertIsNone(result.exception)

    def test_switch_on(self):
        """
        scenario: switch --on/--off detrmines if option opt is required
        action: invoke with --on
        behaviour: fails with Missin option message
        """
        cmd = self.simple_cmd('--on/--off')
        runner = CliRunner()
        result = runner.invoke(cmd, ['--on'])
        self.assertIsNotNone(result.exception)
        self.assertIn('Error: Missing option "--opt".', result.output)

    def test_flag_off(self):
        """
        scenario: flag "--on" detrmines if option opt is required
        action: invoke without options
        behaviour: command runs without complaining
        """
        cmd = self.simple_cmd('--on', is_flag=True)
        runner = CliRunner()
        result = runner.invoke(cmd, [])
        self.assertIsNone(result.exception)

    def test_flag_on(self):
        """
        scenario: flag "--on" detrmines if option opt is required
        action: invoke with --on
        behaviour: fails with Missing option message
        """
        cmd = self.simple_cmd('--on', is_flag=True)
        runner = CliRunner()
        result = runner.invoke(cmd, ['--on'])
        self.assertIsNotNone(result.exception)
        self.assertIn('Error: Missing option "--opt".', result.output)

    def setup_multi_non_eager(self):
        """
        scenario a-or-b:

            * flag a_or_b (--a/--b)
            * opt-a required if a_or_b == True
            * opt-b required if a_or_b == False
        """

        # pylint: disable=no-self-use

        @click.command()
        @click.option('--a/--b', 'a_or_b')
        @click.option('--opt-a', required_fn=lambda c: c.params.get('a_or_b'), cls=ConditionalOption)
        @click.option('--opt-b', required_fn=lambda c: not c.params.get('a_or_b'), cls=ConditionalOption)
        def cmd(a_or_b, opt_a, opt_b):
            """test command for scenario a-or-b"""
            # pylint: disable=unused-argument

            click.echo('{} / {}'.format(opt_a, opt_b))

        runner = CliRunner()
        return runner, cmd

    def test_aa(self):
        """
        scenario = a-or-b
        action: require a, give a (+ reversed order)
        behaviour: command runs
        """
        runner, cmd = self.setup_multi_non_eager()
        result = runner.invoke(cmd, ['--a', '--opt-a=Bla'])
        self.assertIsNone(result.exception)
        self.assertEqual(result.output, 'Bla / None\n')

        result_rev = runner.invoke(cmd, ['--opt-a=Bla', '--a'])
        self.assertIsNone(result_rev.exception)
        self.assertEqual(result_rev.output, 'Bla / None\n')

    def test_ab(self):
        """
        scenario = a-or-b
        action: require a, give b (+ reversed order)
        behaviour: fail, Missing option
        """
        runner, cmd = self.setup_multi_non_eager()
        result = runner.invoke(cmd, ['--a', '--opt-b=Bla'])
        self.assertIsNotNone(result.exception)
        self.assertIn('Error: Missing option "--opt-a".', result.output)

        result_rev = runner.invoke(cmd, ['--opt-b=Bla', '--a'])
        self.assertIsNotNone(result_rev.exception)
        self.assertIn('Error: Missing option "--opt-a".', result_rev.output)

    def test_ba(self):
        """
        scenario = a-or-b
        action: require b, give a (+ reversed order)
        behaviour: fail, Missing option
        """
        runner, cmd = self.setup_multi_non_eager()
        result = runner.invoke(cmd, ['--b', '--opt-a=Bla'])
        self.assertIsNotNone(result.exception)
        self.assertIn('Error: Missing option "--opt-b".', result.output)

        result_rev = runner.invoke(cmd, ['--opt-a=Bla', '--b'])
        self.assertIsNotNone(result_rev.exception)
        self.assertIn('Error: Missing option "--opt-b".', result_rev.output)

    @staticmethod
    def user_callback(_ctx, param, value):
        """
        Testing callback that does not accept 42 and transforms a missing value to -1
        """
        if not value:
            return -1

        if value != 42:
            raise click.BadParameter('invalid', param=param)

        return value

    @staticmethod
    def setup_flag_cond(**kwargs):
        """Set up a command with a flag and a customizable option that depends on it."""

        @click.command()
        @click.option('--flag', is_flag=True)
        @click.option('--opt-a', required_fn=lambda c: c.params.get('flag'), cls=ConditionalOption, **kwargs)
        def cmd(flag, opt_a):
            """ A command with a flag and customizable options that dependon it """
            # pylint: disable=unused-argument

            click.echo('{}'.format(opt_a))

        return cmd

    def test_default(self):
        """Test that the default still gets passed."""
        cmd = self.setup_flag_cond(default='default')
        result_noflag = self.runner.invoke(cmd)
        self.assertIsNone(result_noflag.exception)
        self.assertEqual('default\n', result_noflag.output)

        result_flag = self.runner.invoke(cmd, ['--flag'])
        self.assertIsNone(result_flag.exception)
        self.assertEqual('default\n', result_flag.output)

    def test_callback(self):
        """Test that the callback still gets called."""
        cmd = self.setup_flag_cond(default=23, type=int, callback=self.user_callback)
        result_noflag = self.runner.invoke(cmd)
        self.assertIsNotNone(result_noflag.exception)

        result_flag = self.runner.invoke(cmd, ['--flag'])
        self.assertIsNotNone(result_flag.exception)

    def test_prompt_callback(self):
        """Test that the callback gets called on prompt results."""
        cmd = self.setup_flag_cond(prompt='A', default=23, type=int, callback=self.user_callback)
        result_noflag = self.runner.invoke(cmd, input='\n')
        self.assertIsNotNone(result_noflag.exception)
        self.assertIn('A [23]: \n', result_noflag.output)
        self.assertIn('Invalid', result_noflag.output)

        result_flag = self.runner.invoke(cmd, ['--flag'], input='\n')
        self.assertIsNotNone(result_flag.exception)
        self.assertIn('A [23]: \n', result_flag.output)
        self.assertIn('Invalid', result_flag.output)

    def test_required(self):
        """Test that required_fn overrides required if it evaluates to False."""
        cmd = self.setup_flag_cond(required=True)
        result_noflag = self.runner.invoke(cmd)
        self.assertIsNone(result_noflag.exception)
        result_flag = self.runner.invoke(cmd, ['--flag'])
        self.assertIsNotNone(result_flag.exception)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for the backend implementation of the `AuthInfo` ORM class."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import six

from . import backends

__all__ = ('BackendAuthInfo', 'BackendAuthInfoCollection')


@six.add_metaclass(abc.ABCMeta)
class BackendAuthInfo(backends.BackendEntity):
    """Backend implementation for the `AuthInfo` ORM class."""

    METADATA_WORKDIR = 'workdir'

    @abc.abstractproperty
    def enabled(self):
        """Return whether this instance is enabled.

        :return: boolean, True if enabled, False otherwise
        """

    @enabled.setter
    def enabled(self, value):
        """Set the enabled state

        :param enabled: boolean, True to enable the instance, False to disable it
        """

    @abc.abstractproperty
    def computer(self):
        """Return the computer associated with this instance.

        :return: :class:`aiida.orm.implementation.computers.BackendComputer`
        """

    @abc.abstractproperty
    def user(self):
        """Return the user associated with this instance.

        :return: :class:`aiida.orm.implementation.users.BackendUser`
        """

    @abc.abstractmethod
    def get_auth_params(self):
        """Return the dictionary of authentication parameters

        :return: a dictionary with authentication parameters
        """

    @abc.abstractmethod
    def set_auth_params(self, auth_params):
        """Set the dictionary of authentication parameters

        :param auth_params: a dictionary with authentication parameters
        """

    @abc.abstractmethod
    def get_metadata(self):
        """Return the dictionary of metadata

        :return: a dictionary with metadata
        """

    @abc.abstractmethod
    def set_metadata(self, metadata):
        """Set the dictionary of metadata

        :param metadata: a dictionary with metadata
        """


@six.add_metaclass(abc.ABCMeta)
class BackendAuthInfoCollection(backends.BackendCollection[BackendAuthInfo]):
    """The collection of backend `AuthInfo` entries."""

    ENTITY_CLASS = BackendAuthInfo

    @abc.abstractmethod
    def delete(self, pk):
        """Delete an entry from the collection.

        :param pk: the pk of the entry to delete
        """

    @abc.abstractmethod
    def get(self, computer, user):
        """Return an entry from the collection that is configured for the given computer and user

        :param computer: a :class:`aiida.orm.implementation.computers.BackendComputer` instance
        :param user: a :class:`aiida.orm.implementation.users.BackendUser` instance
        :return: :class:`aiida.orm.implementation.authinfos.BackendAuthInfo`
        :raise aiida.common.exceptions.NotExistent: if no entry exists for the computer/user pair
        :raise aiida.common.exceptions.MultipleObjectsError: if multiple entries exist for the computer/user pair
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Tests for the aiida.common.utils functionality
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import datetime
import unittest

from dateutil.parser import parse

from aiida.common import escaping
from aiida.common import utils


class UniqueTest(unittest.TestCase):
    """
    Tests for the get_unique_filename function.
    """

    def test_unique_1(self):
        filename = "different.txt"
        filename_list = ["file1.txt", "file2.txt"]

        self.assertEqual(filename, utils.get_unique_filename(filename, filename_list))

    def test_unique_2(self):
        filename = "file1.txt"
        filename_list = ["file1.txt", "file2.txt"]

        self.assertEqual("file1-1.txt", utils.get_unique_filename(filename, filename_list))

    def test_unique_3(self):
        filename = "file1.txt"
        filename_list = ["file1.txt", "file1-1.txt"]

        self.assertEqual("file1-2.txt", utils.get_unique_filename(filename, filename_list))

    def test_unique_4(self):
        filename = "file1.txt"
        filename_list = ["file1.txt", "file1-2.txt"]

        self.assertEqual("file1-1.txt", utils.get_unique_filename(filename, filename_list))

    # The counter & the method that increments it and
    # returns its value. It is used in the tests
    seq = -1

    def array_counter(self):
        self.seq += 1
        return self.seq

    def test_query_yes_no(self):
        """
        This method tests the query_yes_no method behaves as expected. To
        perform this, a lambda function is used to simulate the user input.
        """
        from aiida.common.utils import Capturing
        from aiida.manage.backup import backup_utils

        # Capture the sysout for the following code
        with Capturing():
            # Check the yes
            backup_utils.input = lambda _: "y"
            self.assertTrue(backup_utils.query_yes_no("", "yes"))

            backup_utils.input = lambda _: "yes"
            self.assertTrue(backup_utils.query_yes_no("", "yes"))

            # Check the no
            backup_utils.input = lambda _: "no"
            self.assertFalse(backup_utils.query_yes_no("", "yes"))

            backup_utils.input = lambda _: "n"
            self.assertFalse(backup_utils.query_yes_no("", "yes"))

            # Check the empty default value that should
            # lead to an error
            with self.assertRaises(ValueError):
                backup_utils.query_yes_no("", "")

            # Check that a None default value and no answer from
            # the user should lead to the repetition of the query until
            # it is answered properly
            self.seq = -1
            answers = ["", "", "", "yes"]
            backup_utils.input = lambda _: answers[self.array_counter()]
            self.assertTrue(backup_utils.query_yes_no("", None))
            self.assertEqual(self.seq, len(answers) - 1)

            # Check that the default answer is returned
            # when the user doesn't give an answer
            backup_utils.input = lambda _: ""
            self.assertTrue(backup_utils.query_yes_no("", "yes"))

            backup_utils.input = lambda _: ""
            self.assertFalse(backup_utils.query_yes_no("", "no"))

    def test_query_string(self):
        """
        This method tests that the query_string method behaves as expected.
        """
        from aiida.manage.backup import backup_utils

        # None should be returned when empty answer and empty default
        # answer is given
        backup_utils.input = lambda _: ""
        self.assertIsNone(backup_utils.query_string("", ""))

        # If no answer is given then the default answer should be returned
        backup_utils.input = lambda _: ""
        self.assertEqual(backup_utils.query_string("", "Def_answer"), "Def_answer")

        # The answer should be returned when the an answer is given by
        # the user
        backup_utils.input = lambda _: "Usr_answer"
        self.assertEqual(backup_utils.query_string("", "Def_answer"), "Usr_answer")

    def test_ask_backup_question(self):
        """
        This method checks that the combined use of query_string and
        query_yes_no by the ask_backup_question is done as expected.
        """
        from aiida.common.utils import Capturing
        from aiida.manage.backup import backup_utils

        # Capture the sysout for the following code
        with Capturing():
            # Test that a question that asks for an integer is working
            # The given answers are in order:
            # - a non-accepted empty answer
            # - an answer that can not be parsed based on the given type
            # - the final expected answer
            self.seq = -1
            answers = ["", "3fd43", "1", "yes"]
            backup_utils.input = lambda _: answers[self.array_counter()]
            self.assertEqual(backup_utils.ask_question("", int, False), int(answers[2]))

            # Test that a question that asks for a date is working correctly.
            # The behavior is similar to the above test.
            self.seq = -1
            answers = ["", "3fd43", "2015-07-28 20:48:53.197537+02:00", "yes"]
            backup_utils.input = lambda _: answers[self.array_counter()]
            self.assertEqual(backup_utils.ask_question("", datetime.datetime, False), parse(answers[2]))

            # Check that None is not allowed as answer
            question = ""
            answer = ""
            backup_utils.input = lambda x: answer if x == question else "y"
            self.assertEqual(backup_utils.ask_question(question, int, True), None)


class PrettifierTest(unittest.TestCase):
    """
    Tests for the Prettifier class methods.
    """

    def test_prettifier(self):
        """
        Check that the prettified strings work as expected for
        a number of different labels and different codes that should
        show them
        """
        prettifier_data = {
            'agr_seekpath': {
                'DELTA_5': r'\xD\f{}\s5\N',
            },
            'agr_simple': {
                'G': r'\xG',
                'Boo3': r'Boo\s3\N',
                'Boo99': r'Boo\s99\N',
            },
            'latex_simple': {
                'G': r'$\Gamma$',
                'Delta9': r'Delta$_{9}$',
                'Delta90': r'Delta$_{90}$',
            },
            'latex_seekpath': {
                'LAMBDA': r'$\Lambda$',
                'something_2': r'something$_{2}$',
            },
            'gnuplot_simple': {
                'G': u'Γ',
                'bla3': r'bla_{3}',
                'bla33': r'bla_{33}',
            },
            'gnuplot_seekpath': {
                'SIGMA': u'Σ',
                'bla_3': r'bla_{3}',
            },
            'pass': {
                'foo': 'foo',
            },
        }

        for prettifier_id in utils.Prettifier.get_prettifiers():
            prettifier = utils.Prettifier(prettifier_id)

            for label, prettified in prettifier_data[prettifier_id].items():
                self.assertEqual(prettifier.prettify(label), prettified)


class SqlStringMatchTest(unittest.TestCase):
    """"
    Test the functions to convert SQL to regex patterns
    """

    def test_patterns(self):
        """"
        Test complex patterns to see if the logic of conversion is correct
        """
        for pattern, match_true, match_false in [
            (r"aa\_a%a_b", ["aa_abbbaab", "aa_aa_b", "aa_abaa_b", "aa_aaxb", "aa_abaaxb"], ["aaaba_b", "aa_abab_b"]),
            (r"^aa[\_%]b", ["^aa[_aaa]b", "^aa[_]b", "^aa[_1]b"], ["^aa[]b", "aa[1]b", "^aa_b"]),
            (r"z^aa^a\sd[\__%]*sa$dfa&s\%d$a", [r"z^aa^a\sd[_d]*sa$dfa&s%d$a", r"z^aa^a\sd[_aaa]*sa$dfa&s%d$a"],
             [r"z^aa^a\sd[_]*sa$dfa&s%d$a", "zaa^asd[_aa]*sa$a"])
        ]:
            for sample in match_true:
                self.assertTrue(
                    escaping.sql_string_match(string=sample, pattern=pattern),
                    "String '{}' should have matched pattern '{}'".format(sample, pattern))
            for sample in match_false:
                self.assertFalse(
                    escaping.sql_string_match(string=sample, pattern=pattern),
                    "String '{}' should not have matched pattern '{}'".format(sample, pattern))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Unique constraints for the db_dbgroup_dbnodes table

Revision ID: 7a6587e16f4c
Revises: 35d4ee9a1b0e
Create Date: 2019-02-11 19:25:11.744902

"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from alembic import op

# revision identifiers, used by Alembic.
revision = '7a6587e16f4c'
down_revision = '35d4ee9a1b0e'
branch_labels = None
depends_on = None


def upgrade():
    """
    Add unique constraints to the db_dbgroup_dbnodes table.
    """
    op.create_unique_constraint('uix_dbnode_id_dbgroup_id', 'db_dbgroup_dbnodes', ['dbnode_id', 'dbgroup_id'])


def downgrade():
    """
    Remove unique constraints from the db_dbgroup_dbnodes table.
    """
    op.drop_constraint('uix_dbnode_id_dbgroup_id', 'db_dbgroup_dbnodes')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Django settings for the AiiDA project.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import sys
import os

from aiida.common.exceptions import ConfigurationError, MissingConfigurationError

from aiida.backends import settings
from aiida.common.setup import parse_repository_uri
from aiida.manage.configuration import get_config
from aiida.common.timezone import get_current_timezone

# Assumes that parent directory of aiida is root for
# things like templates/, SQL/ etc.  If not, change what follows...

AIIDA_DIR = os.path.dirname(os.path.abspath(__file__))
BASE_DIR = os.path.split(AIIDA_DIR)[0]
sys.path = [BASE_DIR] + sys.path

try:
    CONFIG = get_config()
except MissingConfigurationError:
    raise MissingConfigurationError("Please run the AiiDA Installation, no config found")

if settings.AIIDADB_PROFILE is None:
    raise ConfigurationError("settings.AIIDADB_PROFILE not defined, did you load django"
                             "through the AiiDA load_dbenv()?")

PROFILE = CONFIG.current_profile
PROFILE_CONF = PROFILE.dictionary

DATABASES = {
    'default': {
        # Add 'postgresql_psycopg2', 'mysql', or 'oracle'.
        'ENGINE': 'django.db.backends.' + PROFILE_CONF.get('AIIDADB_ENGINE', ''),
        'NAME': PROFILE_CONF.get('AIIDADB_NAME', ''),
        'USER': PROFILE_CONF.get('AIIDADB_USER', ''),
        'PASSWORD': PROFILE_CONF.get('AIIDADB_PASS', ''),
        'HOST': PROFILE_CONF.get('AIIDADB_HOST', ''),
        'PORT': PROFILE_CONF.get('AIIDADB_PORT', ''),
    }
}

# Checks on the REPOSITORY_* variables
try:
    REPOSITORY_URI = PROFILE_CONF['AIIDADB_REPOSITORY_URI']
except KeyError:
    raise ConfigurationError("Please setup correctly the REPOSITORY_URI variable to "
                             "a suitable directory on which you have write permissions.")

# Note: this variable might disappear in the future
REPOSITORY_PROTOCOL, REPOSITORY_PATH = parse_repository_uri(REPOSITORY_URI)

if settings.IN_RT_DOC_MODE:
    pass
elif REPOSITORY_PROTOCOL == 'file':
    # Note: to verify whether this is too slow to do at every startup
    if not os.path.isdir(REPOSITORY_PATH):
        try:
            # Try to create the local repository folders with needed parent
            # folders
            os.makedirs(REPOSITORY_PATH)
        except OSError:
            # Possibly here due to permission problems
            raise ConfigurationError("Please setup correctly the REPOSITORY_PATH variable to "
                                     "a suitable directory on which you have write permissions. "
                                     "(I was not able to create the directory.)")
else:
    raise ConfigurationError("Only file protocol supported")

# CUSTOM USER CLASS
AUTH_USER_MODEL = 'db.DbUser'

# No secret key defined since we do not use Django to serve HTTP pages
SECRET_KEY = 'placeholder'  # noqa

# Automatic logging configuration for Django is disabled here
# and done for all backends in aiida/__init__.py
LOGGING_CONFIG = None

# Keep DEBUG = False! Otherwise every query is stored in memory
DEBUG = False

ADMINS = []
ALLOWED_HOSTS = []

MANAGERS = ADMINS

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

# Local time zone for this installation. Always choose the system timezone
TIME_ZONE = get_current_timezone().zone

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = False

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale.
USE_L10N = False

# If you set this to False, Django will not use timezone-aware datetimes.
# For AiiDA, leave it as True, otherwise setting properties with dates will not work.
USE_TZ = True

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
            'debug':
            DEBUG,
        },
    },
]

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'aiida.backends.djsite.db',
]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi run` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import contextlib
import os
import sys

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params.options.multivalue import MultipleValueOption
from aiida.cmdline.utils import decorators, echo


@contextlib.contextmanager
def update_environment(new_argv):
    """
    Used as a context manager, changes sys.argv with the
    new_argv argument, and restores it upon exit.
    """
    _argv = sys.argv[:]
    sys.argv = new_argv[:]
    yield

    # Restore old parameters when exiting from the context manager
    sys.argv = _argv


@verdi.command('run', context_settings=dict(ignore_unknown_options=True,))
@click.argument('scriptname', type=click.STRING)
@click.argument('varargs', nargs=-1, type=click.UNPROCESSED)
@click.option('-g', '--group', is_flag=True, default=True, show_default=True, help='Enables the autogrouping')
@click.option('-n', '--group-name', type=click.STRING, required=False, help='Specify the name of the auto group')
@click.option('-e', '--exclude', cls=MultipleValueOption, default=[], help='Exclude these classes from auto grouping')
@click.option(
    '-i', '--include', cls=MultipleValueOption, default=['all'], help='Include these classes from auto grouping')
@click.option(
    '-E',
    '--excludesubclasses',
    cls=MultipleValueOption,
    default=[],
    help='Exclude these classes and their sub classes from auto grouping')
@click.option(
    '-I',
    '--includesubclasses',
    cls=MultipleValueOption,
    default=[],
    help='Include these classes and their sub classes from auto grouping')
@decorators.with_dbenv()
def run(scriptname, varargs, group, group_name, exclude, excludesubclasses, include, includesubclasses):
    # pylint: disable=too-many-arguments,exec-used
    """Execute an AiiDA script."""
    from aiida.cmdline.utils.shell import DEFAULT_MODULES_LIST
    from aiida.orm import autogroup

    # Prepare the environment for the script to be run
    globals_dict = {
        '__builtins__': globals()['__builtins__'],
        '__name__': '__main__',
        '__file__': scriptname,
        '__doc__': None,
        '__package__': None
    }

    # Dynamically load modules (the same of verdi shell) - but in globals_dict, not in the current environment
    for app_mod, model_name, alias in DEFAULT_MODULES_LIST:
        globals_dict['{}'.format(alias)] = getattr(__import__(app_mod, {}, {}, model_name), model_name)

    if group:
        automatic_group_name = group_name
        if automatic_group_name is None:
            from aiida.common import timezone

            automatic_group_name = 'Verdi autogroup on ' + timezone.now().strftime("%Y-%m-%d %H:%M:%S")

        aiida_verdilib_autogroup = autogroup.Autogroup()
        aiida_verdilib_autogroup.set_exclude(exclude)
        aiida_verdilib_autogroup.set_include(include)
        aiida_verdilib_autogroup.set_exclude_with_subclasses(excludesubclasses)
        aiida_verdilib_autogroup.set_include_with_subclasses(includesubclasses)
        aiida_verdilib_autogroup.set_group_name(automatic_group_name)

        # Note: this is also set in the exec environment! This is the intended behavior
        autogroup.current_autogroup = aiida_verdilib_autogroup

    # Initialize the variable here, otherwise we get UnboundLocalError in the finally clause if it fails to open
    handle = None

    try:
        # Here we use a standard open and not io.open, as exec will later fail if passed a unicode type string.
        handle = open(scriptname, 'r')
    except IOError:
        echo.echo_critical("Unable to load file '{}'".format(scriptname))
    else:
        try:
            # Must add also argv[0]
            new_argv = [scriptname] + list(varargs)
            with update_environment(new_argv=new_argv):
                # Add local folder to sys.path
                sys.path.insert(0, os.path.abspath(os.curdir))
                # Compile the script for execution and pass it to exec with the globals_dict
                exec(compile(handle.read(), scriptname, 'exec', dont_inherit=True), globals_dict)  # yapf: disable # pylint: disable=exec-used
        except SystemExit:  # pylint: disable=try-except-raise
            # Script called sys.exit()
            # Re-raise the exception to have the error code properly returned at the end
            raise
    finally:
        if handle:
            handle.close()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,too-few-public-methods
"""Renaming `DbNode.type` to `DbNode.node_type`"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

# Remove when https://github.com/PyCQA/pylint/issues/1931 is fixed
# pylint: disable=no-name-in-module,import-error
from django.db import migrations
from aiida.backends.djsite.db.migrations import upgrade_schema_version

REVISION = '1.0.30'
DOWN_REVISION = '1.0.29'


class Migration(migrations.Migration):
    """Renaming `DbNode.type` to `DbNode.node_type`"""

    dependencies = [
        ('db', '0029_rename_parameter_data_to_dict'),
    ]

    operations = [
        migrations.RenameField(
            model_name='dbnode',
            old_name='type',
            new_name='node_type',
        ),
        upgrade_schema_version(REVISION, DOWN_REVISION)
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
try:
    import ultrajson as json
    from functools import partial

    # double_precision = 15, to replicate what PostgreSQL numerical type is
    # using
    json_dumps = partial(json.dumps, double_precision=15)
    json_loads = partial(json.loads, precise_float=True)
except ImportError:
    from aiida.common import json

    json_dumps = json.dumps
    json_loads = json.loads

import datetime
import re

import six
from alembic import command
from alembic.config import Config
from alembic.runtime.environment import EnvironmentContext
from alembic.script import ScriptDirectory
from dateutil import parser
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker

from aiida.backends import sqlalchemy as sa

ALEMBIC_FILENAME = "alembic.ini"
ALEMBIC_REL_PATH = "migrations"


def flag_modified(instance, key):
    """Wrapper around `sqlalchemy.orm.attributes.flag_modified` to correctly dereference utils.ModelWrapper

    Since SqlAlchemy 1.2.12 (and maybe earlier but not in 1.0.19) the flag_modified function will check that the
    key is actually present in the instance or it will except. If we pass a model instance, wrapped in the ModelWrapper
    the call will raise an InvalidRequestError. In this function that wraps the flag_modified of SqlAlchemy, we
    derefence the model instance if the passed instance is actually wrapped in the ModelWrapper.
    """
    from sqlalchemy.orm.attributes import flag_modified as flag_modified_sqla
    from aiida.orm.implementation.sqlalchemy.utils import ModelWrapper

    if isinstance(instance, ModelWrapper):
        instance = instance._model

    flag_modified_sqla(instance, key)


def recreate_after_fork(engine):
    """
    :param engine: the engine that will be used by the sessionmaker

    Callback called after a fork. Not only disposes the engine, but also recreates a new scoped session
    to use independent sessions in the forked process.
    """
    sa.engine.dispose()
    sa.scopedsessionclass = scoped_session(sessionmaker(bind=sa.engine, expire_on_commit=True))


def reset_session(config):
    """
    :param config: the configuration of the profile from the
       configuration file

    Resets (global) engine and sessionmaker classes, to create a new one
    (or creates a new one from scratch if not already available)
    """
    from multiprocessing.util import register_after_fork

    engine_url = (
        "postgresql://{AIIDADB_USER}:{AIIDADB_PASS}@"
        "{AIIDADB_HOST}{sep}{AIIDADB_PORT}/{AIIDADB_NAME}"
        ).format(sep=':' if config['AIIDADB_PORT'] else '', **config)

    sa.engine = create_engine(engine_url, json_serializer=dumps_json,
                              json_deserializer=loads_json, encoding='utf-8')
    sa.scopedsessionclass = scoped_session(sessionmaker(bind=sa.engine,
                                                        expire_on_commit=True))
    register_after_fork(sa.engine, recreate_after_fork)


def load_dbenv(profile=None, connection=None):
    """
    Load the database environment (SQLAlchemy) and perform some checks.

    :param profile: the string with the profile to use. If not specified,
        use the default one specified in the AiiDA configuration file.
    """
    _load_dbenv_noschemacheck(profile=profile)
    # Check schema version and the existence of the needed tables
    check_schema_version(profile_name=profile)


def _load_dbenv_noschemacheck(profile=None, connection=None):
    """
    Load the SQLAlchemy database.
    """
    from aiida.manage.configuration import get_config

    config = get_config()
    profile = config.current_profile
    reset_session(profile.dictionary)


_aiida_autouser_cache = None


def dumps_json(d):
    """
    Transforms all datetime object into isoformat and then returns the JSON
    """

    def f(v):
        if isinstance(v, list):
            return [f(_) for _ in v]
        elif isinstance(v, dict):
            return dict((key, f(val)) for key, val in v.items())
        elif isinstance(v, datetime.datetime):
            return v.isoformat()
        return v

    return json_dumps(f(d))


date_reg = re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+(\+\d{2}:\d{2})?$')


def loads_json(s):
    """
    Loads the json and try to parse each basestring as a datetime object
    """

    ret = json_loads(s)

    def f(d):
        if isinstance(d, list):
            for i, val in enumerate(d):
                d[i] = f(val)
            return d
        elif isinstance(d, dict):
            for k, v in d.items():
                d[k] = f(v)
            return d
        elif isinstance(d, six.string_types):
            if date_reg.match(d):
                try:
                    return parser.parse(d)
                except (ValueError, TypeError):
                    return d
            return d
        return d

    return f(ret)


# XXX the code here isn't different from the one use in Django. We may be able
# to refactor it in some way
def install_tc(session):
    """
    Install the transitive closure table with SqlAlchemy.
    """
    links_table_name = "db_dblink"
    links_table_input_field = "input_id"
    links_table_output_field = "output_id"
    closure_table_name = "db_dbpath"
    closure_table_parent_field = "parent_id"
    closure_table_child_field = "child_id"

    session.execute(get_pg_tc(links_table_name, links_table_input_field,
                              links_table_output_field, closure_table_name,
                              closure_table_parent_field,
                              closure_table_child_field))


def get_pg_tc(links_table_name,
              links_table_input_field,
              links_table_output_field,
              closure_table_name,
              closure_table_parent_field,
              closure_table_child_field):
    """
    Return the transitive closure table template
    """
    from string import Template

    pg_tc = Template("""

DROP TRIGGER IF EXISTS autoupdate_tc ON $links_table_name;
DROP FUNCTION IF EXISTS update_tc();

CREATE OR REPLACE FUNCTION update_tc()
  RETURNS trigger AS
$$BODY$$
DECLARE

    new_id INTEGER;
    old_id INTEGER;
    num_rows INTEGER;

BEGIN

  IF tg_op = 'INSERT' THEN

    IF EXISTS (
      SELECT Id FROM $closure_table_name
      WHERE $closure_table_parent_field = new.$links_table_input_field
         AND $closure_table_child_field = new.$links_table_output_field
         AND depth = 0
         )
    THEN
      RETURN null;
    END IF;

    IF new.$links_table_input_field = new.$links_table_output_field
    OR EXISTS (
      SELECT id FROM $closure_table_name
        WHERE $closure_table_parent_field = new.$links_table_output_field
        AND $closure_table_child_field = new.$links_table_input_field
        )
    THEN
      RETURN null;
    END IF;

    INSERT INTO $closure_table_name (
         $closure_table_parent_field,
         $closure_table_child_field,
         depth)
      VALUES (
         new.$links_table_input_field,
         new.$links_table_output_field,
         0);

    new_id := lastval();

    UPDATE $closure_table_name
      SET entry_edge_id = new_id
        , exit_edge_id = new_id
        , direct_edge_id = new_id
      WHERE id = new_id;


    INSERT INTO $closure_table_name (
      entry_edge_id,
      direct_edge_id,
      exit_edge_id,
      $closure_table_parent_field,
      $closure_table_child_field,
      depth)
      SELECT id
         , new_id
         , new_id
         , $closure_table_parent_field
         , new.$links_table_output_field
         , depth + 1
        FROM $closure_table_name
        WHERE $closure_table_child_field = new.$links_table_input_field;


    INSERT INTO $closure_table_name (
      entry_edge_id,
      direct_edge_id,
      exit_edge_id,
      $closure_table_parent_field,
      $closure_table_child_field,
      depth)
      SELECT new_id
        , new_id
        , id
        , new.$links_table_input_field
        , $closure_table_child_field
        , depth + 1
        FROM $closure_table_name
        WHERE $closure_table_parent_field = new.$links_table_output_field;

    INSERT INTO $closure_table_name (
      entry_edge_id,
      direct_edge_id,
      exit_edge_id,
      $closure_table_parent_field,
      $closure_table_child_field,
      depth)
      SELECT A.id
        , new_id
        , B.id
        , A.$closure_table_parent_field
        , B.$closure_table_child_field
        , A.depth + B.depth + 2
     FROM $closure_table_name A
        CROSS JOIN $closure_table_name B
     WHERE A.$closure_table_child_field = new.$links_table_input_field
       AND B.$closure_table_parent_field = new.$links_table_output_field;

  END IF;

  IF tg_op = 'DELETE' THEN

    IF NOT EXISTS(
        SELECT id FROM $closure_table_name
        WHERE $closure_table_parent_field = old.$links_table_input_field
        AND $closure_table_child_field = old.$links_table_output_field AND
        depth = 0 )
    THEN
        RETURN NULL;
    END IF;

    CREATE TABLE PurgeList (Id int);

    INSERT INTO PurgeList
      SELECT id FROM $closure_table_name
          WHERE $closure_table_parent_field = old.$links_table_input_field
        AND $closure_table_child_field = old.$links_table_output_field AND
        depth = 0;

    WHILE (1 = 1)
    loop

      INSERT INTO PurgeList
        SELECT id FROM $closure_table_name
          WHERE depth > 0
          AND ( entry_edge_id IN ( SELECT Id FROM PurgeList )
          OR direct_edge_id IN ( SELECT Id FROM PurgeList )
          OR exit_edge_id IN ( SELECT Id FROM PurgeList ) )
          AND Id NOT IN (SELECT Id FROM PurgeList );

      GET DIAGNOSTICS num_rows = ROW_COUNT;
      if (num_rows = 0) THEN
        EXIT;
      END IF;
    end loop;

    DELETE FROM $closure_table_name WHERE Id IN ( SELECT Id FROM PurgeList);
    DROP TABLE PurgeList;

  END IF;

  RETURN NULL;

END
$$BODY$$
  LANGUAGE plpgsql VOLATILE
  COST 100;


CREATE TRIGGER autoupdate_tc
  AFTER INSERT OR DELETE OR UPDATE
  ON $links_table_name FOR each ROW
  EXECUTE PROCEDURE update_tc();

""")
    return pg_tc.substitute(links_table_name=links_table_name,
                            links_table_input_field=links_table_input_field,
                            links_table_output_field=links_table_output_field,
                            closure_table_name=closure_table_name,
                            closure_table_parent_field=closure_table_parent_field,
                            closure_table_child_field=closure_table_child_field)


def migrate_database(alembic_cfg=None):
    """Migrate the database to the latest schema version.

    :param config: alembic configuration to use, will use default if not provided
    """
    from aiida.backends import sqlalchemy as sa

    if alembic_cfg is None:
        alembic_cfg = get_alembic_conf()

    with sa.engine.connect() as connection:
        alembic_cfg.attributes['connection'] = connection
        command.upgrade(alembic_cfg, "head")


def check_schema_version(profile_name=None):
    """
    Check if the version stored in the database is the same of the version of the code.

    :raise aiida.common.ConfigurationError: if the two schema versions do not match
    """
    from aiida.backends import sqlalchemy as sa
    from aiida.backends.settings import IN_DOC_MODE
    from aiida.common.exceptions import ConfigurationError

    # Early exit if we compile the documentation since the schema check is not needed and it creates problems
    # with the sqlalchemy migrations
    if IN_DOC_MODE:
        return

    alembic_cfg = get_alembic_conf()

    # Getting the version of the code and the database, reusing the existing engine (initialized by AiiDA)
    with sa.engine.begin() as connection:
        alembic_cfg.attributes['connection'] = connection
        code_schema_version = get_migration_head(alembic_cfg)
        db_schema_version = get_db_schema_version(alembic_cfg)

    if code_schema_version != db_schema_version:
        from aiida.manage.manager import get_manager
        manager = get_manager()
        profile_name = manager.get_profile().name
        raise ConfigurationError('Database schema version {} is outdated compared to the code schema version {}\n'
                                 'To migrate the database to the current version, run the following commands:'
                                 '\n  verdi -p {} daemon stop\n  verdi -p {} database migrate'.format(
                                    code_schema_version, db_schema_version, profile_name, profile_name))



def get_migration_head(config):
    """
    This function returns the head of the migration scripts.
    :param config: The alembic configuration.
    :return: The version of the head.
    """
    script = ScriptDirectory.from_config(config)
    return script.get_current_head()


def get_db_schema_version(config):
    """
    This function returns the current version of the database.
    :param config: The alembic configuration.
    :return: The version of the database.
    """
    if config is None:
        return None

    script = ScriptDirectory.from_config(config)

    def get_db_version(rev, _):
        if isinstance(rev, tuple) and len(rev) > 0:
            config.attributes['rev'] = rev[0]
        else:
            config.attributes['rev'] = None

        return []

    with EnvironmentContext(
            config,
            script,
            fn=get_db_version
    ):
        script.run_env()
        return config.attributes['rev']


def get_alembic_conf():
    """
    This function returns the alembic configuration file contents by doing
    the necessary updates in the 'script_location' name.
    :return: The alembic configuration.
    """
    # Constructing the alembic full path & getting the configuration
    import os
    dir_path = os.path.dirname(os.path.realpath(__file__))
    alembic_fpath = os.path.join(dir_path, ALEMBIC_FILENAME)
    alembic_cfg = Config(alembic_fpath)

    # Set the alembic script directory location
    alembic_dpath = os.path.join(dir_path, ALEMBIC_REL_PATH)
    alembic_cfg.set_main_option('script_location', alembic_dpath)

    return alembic_cfg


def delete_nodes_and_connections_sqla(pks_to_delete):
    """
    Delete all nodes corresponding to pks in the input.
    :param pks_to_delete: A list, tuple or set of pks that should be deleted.
    """
    from aiida.backends import sqlalchemy as sa
    from aiida.backends.sqlalchemy.models.node import DbNode, DbLink
    from aiida.backends.sqlalchemy.models.group import table_groups_nodes

    session = sa.get_scoped_session()
    try:
        # I am first making a statement to delete the membership of these nodes to groups.
        # Since table_groups_nodes is a sqlalchemy.schema.Table, I am using expression language to compile
        # a stmt to be executed by the session. It works, but it's not nice that two different ways are used!
        # Can this be changed?
        stmt = table_groups_nodes.delete().where(table_groups_nodes.c.dbnode_id.in_(list(pks_to_delete)))
        session.execute(stmt)
        # First delete links, then the Nodes, since we are not cascading deletions.
        # Here I delete the links coming out of the nodes marked for deletion.
        session.query(DbLink).filter(DbLink.input_id.in_(list(pks_to_delete))).delete(synchronize_session='fetch')
        # Here I delete the links pointing to the nodes marked for deletion.
        session.query(DbLink).filter(DbLink.output_id.in_(list(pks_to_delete))).delete(synchronize_session='fetch')
        # Now I am deleting the nodes
        session.query(DbNode).filter(DbNode.id.in_(list(pks_to_delete))).delete(synchronize_session='fetch')
        # Here I commit this scoped session!
        session.commit()
    except Exception as e:
        # If there was any exception, I roll back the session.
        session.rollback()
        raise e
    finally:
        session.close()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""AiiDA specific implementation of plumpy's ProcessSpec."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import six
import plumpy

from .exit_code import ExitCode, ExitCodesNamespace
from .ports import InputPort, PortNamespace, CalcJobOutputPort


class ProcessSpec(plumpy.ProcessSpec):
    """Default process spec for process classes defined in `aiida-core`.

    This sub class defines custom classes for input ports and port namespaces. It also adds support for the definition
    of exit codes and retrieving them subsequently.
    """

    METADATA_KEY = 'metadata'
    METADATA_OPTIONS_KEY = 'options'
    INPUT_PORT_TYPE = InputPort
    PORT_NAMESPACE_TYPE = PortNamespace

    def __init__(self):
        super(ProcessSpec, self).__init__()
        self._exit_codes = ExitCodesNamespace()

    @property
    def metadata_key(self):
        return self.METADATA_KEY

    @property
    def options_key(self):
        return self.METADATA_OPTIONS_KEY

    @property
    def exit_codes(self):
        """
        Return the namespace of exit codes defined for this ProcessSpec

        :returns: ExitCodesNamespace of ExitCode named tuples
        """
        return self._exit_codes

    def exit_code(self, status, label, message):
        """
        Add an exit code to the ProcessSpec

        :param status: the exit status integer
        :param label: a label by which the exit code can be addressed
        :param message: a more detailed description of the exit code
        """
        if not isinstance(status, int):
            raise TypeError('status should be of integer type and not of {}'.format(type(status)))

        if status < 0:
            raise ValueError('status should be a positive integer, received {}'.format(type(status)))

        if not isinstance(label, six.string_types):
            raise TypeError('label should be of basestring type and not of {}'.format(type(label)))

        if not isinstance(message, six.string_types):
            raise TypeError('message should be of basestring type and not of {}'.format(type(message)))

        self._exit_codes[label] = ExitCode(status, message)


class CalcJobProcessSpec(ProcessSpec):
    """Process spec intended for the `CalcJob` process class."""

    OUTPUT_PORT_TYPE = CalcJobOutputPort

    def __init__(self):
        super(CalcJobProcessSpec, self).__init__()
        self._default_output_node = None

    @property
    def default_output_node(self):
        return self._default_output_node

    @default_output_node.setter
    def default_output_node(self, port_name):
        from aiida.orm import Dict

        if port_name not in self.outputs:
            raise ValueError('{} is not a registered output port'.format(port_name))

        valid_type_port = self.outputs[port_name].valid_type
        valid_type_required = Dict

        if valid_type_port is not valid_type_required:
            raise ValueError('the valid type of a default output has to be a {} but it is {}'.format(
                valid_type_port, valid_type_required))

        self._default_output_node = port_name



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tools to operate on `CifData` nodes."""
# pylint: disable=invalid-name
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from six.moves import range

from aiida.orm import CifData
from aiida.engine import calcfunction


class InvalidOccupationsError(Exception):
    """
    An exception that will be raised if pymatgen fails to parse the structure from a
    cif because some site occupancies exceed the occupancy tolerance. This often happens
    for structures that have attached species, such as hydrogen, and specify a placeholder
    position for it, leading to occupancies greater than one. Pymatgen only issues a
    warning in this case and simply does not return a structure
    """


symmetry_tags = [
    '_symmetry_equiv_pos_site_id',
    '_symmetry_equiv_pos_as_xyz',
    '_symmetry_Int_Tables_number',
    '_symmetry_space_group_name_H-M',
    '_symmetry_space_group_name_Hall',
    '_space_group_symop_id',
    '_space_group_symop_operation_xyz',
    '_space_group_symop_sg_id',
    '_space_group_id',
    '_space_group_IT_number',
    '_space_group_name_H-M_alt',
    '_space_group_name_Hall',
]


def symop_string_from_symop_matrix_tr(matrix, tr=(0, 0, 0), eps=0):
    """
    Construct a CIF representation of symmetry operator plus translation.
    See International Tables for Crystallography Vol. A. (2002) for
    definition.

    :param matrix: 3x3 matrix, representing the symmetry operator
    :param tr: translation vector of length 3 (default 0)
    :param eps: epsilon parameter for fuzzy comparison x == 0
    :return: CIF representation of symmetry operator
    """
    import re
    axes = ["x", "y", "z"]
    parts = ["", "", ""]
    for i in range(3):
        for j in range(3):
            sign = None
            if matrix[i][j] > eps:
                sign = "+"
            elif matrix[i][j] < -eps:
                sign = "-"
            if sign:
                parts[i] = format("{}{}{}".format(parts[i], sign, axes[j]))
        if tr[i] < -eps or tr[i] > eps:
            sign = "+"
            if tr[i] < -eps:
                sign = "-"
            parts[i] = format("{}{}{}".format(parts[i], sign, abs(tr[i])))
        parts[i] = re.sub(r'^\+', '', parts[i])
    return ",".join(parts)


@calcfunction
def _get_aiida_structure_ase_inline(cif, **kwargs):
    """
    Creates :py:class:`aiida.orm.nodes.data.structure.StructureData` using ASE.

    .. note:: unable to correctly import structures of alloys.
    .. note:: requires ASE module.
    """
    from aiida.orm import Dict, StructureData

    parameters = kwargs.get('parameters', {})

    if isinstance(parameters, Dict):
        parameters = parameters.get_dict()

    parameters.pop('occupancy_tolerance', None)
    parameters.pop('site_tolerance', None)

    return {'structure': StructureData(ase=cif.get_ase(**parameters))}


@calcfunction
def _get_aiida_structure_pymatgen_inline(cif, **kwargs):
    """
    Creates :py:class:`aiida.orm.nodes.data.structure.StructureData` using pymatgen.

    :param occupancy_tolerance: If total occupancy of a site is between 1 and occupancy_tolerance,
        the occupancies will be scaled down to 1.
    :param site_tolerance: This tolerance is used to determine if two sites are sitting in the same position,
        in which case they will be combined to a single disordered site. Defaults to 1e-4.

    .. note:: requires pymatgen module.
    """
    from pymatgen.io.cif import CifParser
    from aiida.orm import Dict, StructureData

    parameters = kwargs.get('parameters', {})

    if isinstance(parameters, Dict):
        parameters = parameters.get_dict()

    constructor_kwargs = {}

    parameters['primitive'] = parameters.pop('primitive_cell', False)

    for argument in ['occupancy_tolerance', 'site_tolerance']:
        if argument in parameters:
            constructor_kwargs[argument] = parameters.pop(argument)

    with cif.open() as handle:
        parser = CifParser(handle, **constructor_kwargs)

    try:
        structures = parser.get_structures(**parameters)
    except ValueError:

        # Verify whether the failure was due to wrong occupancy numbers
        try:
            constructor_kwargs['occupancy_tolerance'] = 1E10
            with cif.open() as handle:
                parser = CifParser(handle, **constructor_kwargs)
            structures = parser.get_structures(**parameters)
        except ValueError:
            # If it still fails, the occupancies were not the reason for failure
            raise ValueError('pymatgen failed to provide a structure from the cif file')
        else:
            # If it now succeeds, non-unity occupancies were the culprit
            raise InvalidOccupationsError(
                'detected atomic sites with an occupation number larger than the occupation tolerance')

    return {'structure': StructureData(pymatgen_structure=structures[0])}


@calcfunction
def refine_inline(node):
    """
    Refine (reduce) the cell of :py:class:`aiida.orm.nodes.data.cif.CifData`,
    find and remove symmetrically equivalent atoms.

    :param node: a :py:class:`aiida.orm.nodes.data.cif.CifData` instance.
    :return: dict with :py:class:`aiida.orm.nodes.data.cif.CifData`

    .. note:: can be used as inline calculation.
    """
    from aiida.orm.nodes.data.structure import StructureData, ase_refine_cell

    if len(node.values.keys()) > 1:
        raise ValueError("CifData seems to contain more than one data "
                         "block -- multiblock CIF files are not "
                         "supported yet")

    name = list(node.values.keys())[0]

    original_atoms = node.get_ase(index=None)
    if len(original_atoms) > 1:
        raise ValueError("CifData seems to contain more than one crystal "
                         "structure -- such refinement is not supported "
                         "yet")

    original_atoms = original_atoms[0]

    refined_atoms, symmetry = ase_refine_cell(original_atoms)

    cif = CifData(ase=refined_atoms)
    if name != str(0):
        cif.values.rename(str(0), name)

    # Remove all existing symmetry tags before overwriting:
    for tag in symmetry_tags:
        cif.values[name].RemoveCifItem(tag)

    cif.values[name]['_symmetry_space_group_name_H-M'] = symmetry['hm']
    cif.values[name]['_symmetry_space_group_name_Hall'] = symmetry['hall']
    cif.values[name]['_symmetry_Int_Tables_number'] = symmetry['tables']
    cif.values[name]['_symmetry_equiv_pos_as_xyz'] = \
        [symop_string_from_symop_matrix_tr(symmetry['rotations'][i],
                                           symmetry['translations'][i])
         for i in range(len(symmetry['rotations']))]

    # Summary formula has to be calculated from non-reduced set of atoms.
    cif.values[name]['_chemical_formula_sum'] = \
        StructureData(ase=original_atoms).get_formula(mode='hill',
                                                      separator=' ')

    # If the number of reduced atoms multiplies the number of non-reduced
    # atoms, the new Z value can be calculated.
    if '_cell_formula_units_Z' in node.values[name].keys():
        old_Z = node.values[name]['_cell_formula_units_Z']
        if len(original_atoms) % len(refined_atoms):
            new_Z = old_Z * len(original_atoms) // len(refined_atoms)
            cif.values[name]['_cell_formula_units_Z'] = new_Z

    return {'cif': cif}



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Backend group module"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import abc
import six

from . import backends

__all__ = ('BackendLog', 'BackendLogCollection')


@six.add_metaclass(abc.ABCMeta)
class BackendLog(backends.BackendEntity):
    """
    Backend Log interface
    """

    @abc.abstractproperty
    def uuid(self):
        """
        Get the UUID of the log entry

        :return: The entry's UUID
        :rtype: uuid.UUID
        """

    @abc.abstractproperty
    def time(self):
        """
        Get the time corresponding to the entry

        :return: The entry timestamp
        :rtype: :class:`!datetime.datetime`
        """

    @abc.abstractproperty
    def loggername(self):
        """
        The name of the logger that created this entry

        :return: The entry loggername
        :rtype: basestring
        """

    @abc.abstractproperty
    def levelname(self):
        """
        The name of the log level

        :return: The entry log level name
        :rtype: basestring
        """

    @abc.abstractproperty
    def dbnode_id(self):
        """
        Get the id of the object that created the log entry

        :return: The id of the object that created the log entry
        :rtype: int
        """

    @abc.abstractproperty
    def message(self):
        """
        Get the message corresponding to the entry

        :return: The entry message
        :rtype: basestring
        """

    @abc.abstractproperty
    def metadata(self):
        """
        Get the metadata corresponding to the entry

        :return: The entry metadata
        :rtype: :class:`!json.json`
        """


@six.add_metaclass(abc.ABCMeta)
class BackendLogCollection(backends.BackendCollection[BackendLog]):
    """The collection of Log entries."""

    ENTITY_CLASS = BackendLog

    @abc.abstractmethod
    def delete(self, log_id):
        """
        Remove a Log entry from the collection with the given id

        :param log_id: id of the log to delete
        """

    @abc.abstractmethod
    def delete_many(self, filters):
        """
        Delete multiple log entries in the table
        """



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Unit tests for the AuthInfo ORM class."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.testbase import AiidaTestCase
from aiida.common import exceptions
from aiida.orm import authinfos


class TestAuthinfo(AiidaTestCase):
    """Unit tests for the AuthInfo ORM class."""

    def setUp(self):
        super(TestAuthinfo, self).setUp()
        for auth_info in authinfos.AuthInfo.objects.all():
            authinfos.AuthInfo.objects.delete(auth_info.pk)

        self.auth_info = self.computer.configure()  # pylint: disable=no-member

    def test_set_auth_params(self):
        """Test the auth_params setter."""
        auth_params = {'safe_interval': 100}

        self.auth_info.set_auth_params(auth_params)
        self.assertEqual(self.auth_info.get_auth_params(), auth_params)

    def test_delete(self):
        """Test deleting a single AuthInfo."""
        pk = self.auth_info.pk

        self.assertEqual(len(authinfos.AuthInfo.objects.all()), 1)
        authinfos.AuthInfo.objects.delete(pk)
        self.assertEqual(len(authinfos.AuthInfo.objects.all()), 0)

        with self.assertRaises(exceptions.NotExistent):
            authinfos.AuthInfo.objects.delete(pk)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the `SinglefileData` class."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import tempfile
import io

from aiida.backends.testbase import AiidaTestCase
from aiida.orm import SinglefileData, load_node


class TestSinglefileData(AiidaTestCase):
    """Tests for the `SinglefileData` class."""

    def test_reload_singlefile_data(self):
        """Test writing and reloading a `SinglefileData` instance."""
        content_original = 'some text ABCDE'

        with tempfile.NamedTemporaryFile(mode='w+') as handle:
            filepath = handle.name
            basename = os.path.basename(filepath)
            handle.write(content_original)
            handle.flush()
            node = SinglefileData(file=filepath)

        uuid = node.uuid

        with node.open() as handle:
            content_written = handle.read()

        self.assertEqual(node.list_object_names(), [basename])
        self.assertEqual(content_written, content_original)

        node.store()

        with node.open() as handle:
            content_stored = handle.read()

        self.assertEqual(content_stored, content_original)
        self.assertEqual(node.list_object_names(), [basename])

        node_loaded = load_node(uuid)
        self.assertTrue(isinstance(node_loaded, SinglefileData))

        with node.open() as handle:
            content_loaded = handle.read()

        self.assertEqual(content_loaded, content_original)
        self.assertEqual(node_loaded.list_object_names(), [basename])

        with node_loaded.open() as handle:
            self.assertEqual(handle.read(), content_original)

    def test_construct_from_filelike(self):
        """Test constructing an instance from filelike instead of filepath."""
        content_original = u'some testing text\nwith a newline'

        with tempfile.NamedTemporaryFile(mode='w+') as handle:
            basename = os.path.basename(handle.name)
            handle.write(content_original)
            handle.flush()
            handle.seek(0)
            node = SinglefileData(file=handle)

        with node.open() as handle:
            content_stored = handle.read()

        self.assertEqual(content_stored, content_original)
        self.assertEqual(node.list_object_names(), [basename])

        node.store()

        with node.open() as handle:
            content_stored = handle.read()

        self.assertEqual(content_stored, content_original)
        self.assertEqual(node.list_object_names(), [basename])

    def test_construct_from_string(self):
        """Test constructing an instance from a string."""
        content_original = u'some testing text\nwith a newline'

        with io.StringIO(content_original) as handle:
            node = SinglefileData(file=handle)

        with node.open() as handle:
            content_stored = handle.read()

        self.assertEqual(content_stored, content_original)
        self.assertEqual(node.list_object_names(), [SinglefileData.DEFAULT_FILENAME])

        node.store()

        with node.open() as handle:
            content_stored = handle.read()

        self.assertEqual(content_stored, content_original)
        self.assertEqual(node.list_object_names(), [SinglefileData.DEFAULT_FILENAME])



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import with_statement
from __future__ import absolute_import

from alembic import context
from aiida.backends.settings import IN_DOC_MODE

# The available SQLAlchemy tables
from aiida.backends.sqlalchemy.models.authinfo import DbAuthInfo
from aiida.backends.sqlalchemy.models.comment import DbComment
from aiida.backends.sqlalchemy.models.computer import DbComputer
from aiida.backends.sqlalchemy.models.group import DbGroup
from aiida.backends.sqlalchemy.models.log import DbLog
from aiida.backends.sqlalchemy.models.node import DbComputer, DbLink, DbNode
from aiida.backends.sqlalchemy.models.settings import DbSetting
from aiida.backends.sqlalchemy.models.user import DbUser
from aiida.common.exceptions import DbContentError
from aiida.backends.sqlalchemy.models.base import Base
target_metadata = Base.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    raise NotImplementedError("This feature is not currently supported.")


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    # This is the Alembic Config object, which provides
    # access to the values within the .ini file in use.
    # It is initialized by alembic and we enrich it here
    # to point to the right database configuration.
    config = context.config

    connectable = config.attributes.get('connection', None)

    if connectable is None:
        from aiida.common.exceptions import ConfigurationError
        raise ConfigurationError("An initialized connection is expected "
                                 "for the AiiDA online migrations.")

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            transaction_per_migration=True,
        )

        with context.begin_transaction():
            context.run_migrations()


if not IN_DOC_MODE:
    if context.is_offline_mode():
        run_migrations_offline()
    else:
        run_migrations_online()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This is a mock of a django settings to make the documentation compile without the call to load_dbenv()

For more information on this file, see
https://docs.djangoproject.com/en/1.7/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.7/ref/settings/
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os


BASE_DIR = os.path.dirname(os.path.dirname(__file__))

# No secret key defined since we do not use Django to serve HTTP pages
SECRET_KEY = 'placeholder'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []


# Application definition
INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'pippo.urls'

WSGI_APPLICATION = 'pippo.wsgi.application'


# Database: https://docs.djangoproject.com/en/1.7/ref/settings/#databases
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

# Internationalization: https://docs.djangoproject.com/en/1.7/topics/i18n/
LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images): https://docs.djangoproject.com/en/1.7/howto/static-files/
STATIC_URL = '/static/'



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# yapf: disable
"""Module with pre-defined reusable commandline options that can be used as `click` decorators."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import click

from aiida.backends.profile import BACKEND_DJANGO, BACKEND_SQLA
from .. import types
from .multivalue import MultipleValueOption
from .overridable import OverridableOption

__all__ = (
    'PROFILE', 'CALCULATION', 'CALCULATIONS', 'CODE', 'CODES', 'COMPUTER', 'COMPUTERS', 'DATUM', 'DATA', 'GROUP',
    'GROUPS', 'NODE', 'NODES', 'FORCE', 'SILENT', 'VISUALIZATION_FORMAT', 'INPUT_FORMAT', 'EXPORT_FORMAT',
    'ARCHIVE_FORMAT', 'NON_INTERACTIVE', 'DRY_RUN', 'USER_EMAIL', 'USER_FIRST_NAME', 'USER_LAST_NAME',
    'USER_INSTITUTION', 'BACKEND', 'DB_HOST', 'DB_PORT', 'DB_USERNAME', 'DB_PASSWORD', 'DB_NAME', 'REPOSITORY_PATH',
    'PROFILE_ONLY_CONFIG', 'PROFILE_SET_DEFAULT', 'PREPEND_TEXT', 'APPEND_TEXT', 'LABEL', 'DESCRIPTION', 'INPUT_PLUGIN',
    'CALC_JOB_STATE', 'PROCESS_STATE', 'EXIT_STATUS', 'FAILED', 'LIMIT', 'PROJECT', 'ORDER_BY', 'PAST_DAYS',
    'OLDER_THAN', 'ALL', 'ALL_STATES', 'ALL_USERS', 'GROUP_CLEAR', 'RAW', 'HOSTNAME', 'TRANSPORT', 'SCHEDULER', 'USER',
    'PORT', 'FREQUENCY', 'VERBOSE', 'TIMEOUT', 'FORMULA_MODE', 'TRAJECTORY_INDEX', 'WITH_ELEMENTS',
    'WITH_ELEMENTS_EXCLUSIVE'
)


def valid_process_states():
    """Return a list of valid values for the ProcessState enum."""
    from plumpy import ProcessState
    return tuple(state.value for state in ProcessState)


def valid_calc_job_states():
    """Return a list of valid values for the CalcState enum."""
    from aiida.common.datastructures import CalcJobState
    return tuple(state.value for state in CalcJobState)


def active_process_states():
    """Return a list of process states that are considered active."""
    from plumpy import ProcessState
    return ([
        ProcessState.CREATED.value,
        ProcessState.WAITING.value,
        ProcessState.RUNNING.value,
    ])


PROFILE = OverridableOption(
    '-p', '--profile', 'profile',
    type=types.ProfileParamType(),
    help='Execute the command for this profile instead of the default profile.')

CALCULATION = OverridableOption(
    '-C', '--calculation', 'calculation',
    type=types.CalculationParamType(),
    help='A single calculation identified by its ID or UUID.')

CALCULATIONS = OverridableOption(
    '-C', '--calculations', 'calculations',
    type=types.CalculationParamType(), cls=MultipleValueOption,
    help='One or multiple calculations identified by their ID or UUID.')

CODE = OverridableOption(
    '-X', '--code', 'code',
    type=types.CodeParamType(),
    help='A single code identified by its ID, UUID or label.')

CODES = OverridableOption(
    '-X', '--codes', 'codes',
    type=types.CodeParamType(), cls=MultipleValueOption,
    help='One or multiple codes identified by their ID, UUID or label.')

COMPUTER = OverridableOption(
    '-Y', '--computer', 'computer',
    type=types.ComputerParamType(),
    help='A single computer identified by its ID, UUID or label.')

COMPUTERS = OverridableOption(
    '-Y', '--computers', 'computers',
    type=types.ComputerParamType(), cls=MultipleValueOption,
    help='One or multiple computers identified by their ID, UUID or label.')

DATUM = OverridableOption(
    '-D', '--datum', 'datum',
    type=types.DataParamType(),
    help='A single datum identified by its ID, UUID or label.')

DATA = OverridableOption(
    '-D', '--data', 'data',
    type=types.DataParamType(), cls=MultipleValueOption,
    help='One or multiple data identified by their ID, UUID or label.')

GROUP = OverridableOption(
    '-G', '--group', 'group',
    type=types.GroupParamType(),
    help='A single group identified by its ID, UUID or label.')

GROUPS = OverridableOption(
    '-G', '--groups', 'groups',
    type=types.GroupParamType(), cls=MultipleValueOption,
    help='One or multiple groups identified by their ID, UUID or label.')

NODE = OverridableOption(
    '-N', '--node', 'node',
    type=types.NodeParamType(),
    help='A single node identified by its ID or UUID.')

NODES = OverridableOption(
    '-N', '--nodes', 'nodes',
    type=types.NodeParamType(), cls=MultipleValueOption,
    help='One or multiple nodes identified by their ID or UUID.')

FORCE = OverridableOption(
    '-f', '--force',
    is_flag=True, default=False,
    help='Do not ask for confirmation.')

SILENT = OverridableOption(
    '-s', '--silent',
    is_flag=True, default=False,
    help='Suppress any output printed to stdout.')

VISUALIZATION_FORMAT = OverridableOption(
    '-F', '--format', 'fmt', show_default=True, help='Format of the visualized output.')

INPUT_FORMAT = OverridableOption(
    '-F', '--format', 'fmt', show_default=True, help='Format of the input file.')

EXPORT_FORMAT = OverridableOption(
    '-F', '--format', 'fmt', show_default=True, help='Format of the exported file.')

ARCHIVE_FORMAT = OverridableOption(
    '-F', '--archive-format',
    type=click.Choice(['zip', 'zip-uncompressed', 'tar.gz']), default='zip', show_default=True,
    help='The format of the archive file.')

NON_INTERACTIVE = OverridableOption(
    '-n', '--non-interactive',
    is_flag=True, is_eager=True,
    help='Non-interactive mode: never prompt for input.')

DRY_RUN = OverridableOption(
    '-n', '--dry-run', is_flag=True,
    help='Perform a dry run.')

USER_EMAIL = OverridableOption(
    '--email',
    type=click.STRING, prompt='Email Address (identifies your data when sharing)',
    help='Email address that will be associated with your data and will be exported along with it, '
    'should you choose to share any of your work.')

USER_FIRST_NAME = OverridableOption(
    '--first-name',
    type=click.STRING, prompt='First name',
    help='First name of the user.')

USER_LAST_NAME = OverridableOption(
    '--last-name',
    type=click.STRING, prompt='Last name',
    help='Last name of the user.')

USER_INSTITUTION = OverridableOption(
    '--institution',
    type=click.STRING, prompt='Institution',
    help='Institution name of the user.')

BACKEND = OverridableOption(
    '--backend',
    type=click.Choice([BACKEND_DJANGO, BACKEND_SQLA]), default=BACKEND_DJANGO,
    help='Database backend to use.')

DB_HOST = OverridableOption(
    '--db-host',
    type=click.STRING,
    help='Database server host.')

DB_PORT = OverridableOption(
    '--db-port',
    type=click.INT,
    help='Database server port.')

DB_USERNAME = OverridableOption(
    '--db-username',
    type=click.STRING,
    help='Database user name.')

DB_PASSWORD = OverridableOption(
    '--db-password',
    type=click.STRING,
    help='Database user password.')

DB_NAME = OverridableOption(
    '--db-name',
    type=click.STRING,
    help='Database name.')

REPOSITORY_PATH = OverridableOption(
    '--repository',
    type=click.Path(file_okay=False),
    help='Absolute path for the file system repository.')

PROFILE_ONLY_CONFIG = OverridableOption(
    '--only-config',
    is_flag=True, default=False,
    help='Only configure the user and skip creating the database.')

PROFILE_SET_DEFAULT = OverridableOption(
    '--set-default',
    is_flag=True, default=False,
    help='Set the profile as the new default.')

PREPEND_TEXT = OverridableOption(
    '--prepend-text',
    type=click.STRING, default='',
    help='Bash script to be executed before an action.')

APPEND_TEXT = OverridableOption(
    '--append-text',
    type=click.STRING, default='',
    help='Bash script to be executed after an action has completed.')

LABEL = OverridableOption(
    '-L', '--label',
    type=click.STRING, metavar='LABEL',
    help='Short name to be used as a label.')

DESCRIPTION = OverridableOption(
    '-D', '--description',
    type=click.STRING, metavar='DESCRIPTION', default='', required=False,
    help='A detailed description.')

INPUT_PLUGIN = OverridableOption(
    '-P', '--input-plugin',
    type=types.PluginParamType(group='calculations'),
    help='Calculation input plugin string.')

CALC_JOB_STATE = OverridableOption(
    '-s', '--calc-job-state', 'calc_job_state',
    type=types.LazyChoice(valid_calc_job_states), cls=MultipleValueOption,
    help='Only include entries with this calculation job state.')

PROCESS_STATE = OverridableOption(
    '-S', '--process-state', 'process_state',
    type=types.LazyChoice(valid_process_states), cls=MultipleValueOption, default=active_process_states,
    help='Only include entries with this process state.')

EXIT_STATUS = OverridableOption(
    '-E', '--exit-status', 'exit_status',
    type=click.INT,
    help='Only include entries with this exit status.')

FAILED = OverridableOption(
    '-X', '--failed', 'failed',
    is_flag=True, default=False,
    help='Only include entries that have failed.')

LIMIT = OverridableOption(
    '-l', '--limit', 'limit',
    type=click.INT, default=None,
    help='Limit the number of entries to display.')

PROJECT = OverridableOption(
    '-P', '--project', 'project',
    cls=MultipleValueOption,
    help='Select the list of entity attributes to project.')

ORDER_BY = OverridableOption(
    '-O', '--order-by', 'order_by',
    type=click.Choice(['id', 'ctime']), default='ctime', show_default=True,
    help='Order the entries by this attribute.')

PAST_DAYS = OverridableOption(
    '-p', '--past-days', 'past_days',
    type=click.INT, metavar='PAST_DAYS',
    help='Only include entries created in the last PAST_DAYS number of days.')

OLDER_THAN = OverridableOption(
    '-o', '--older-than', 'older_than',
    type=click.INT, metavar='OLDER_THAN',
    help='Only include entries created before OLDER_THAN days ago.')

ALL = OverridableOption(
    '-a', '--all', 'all_entries',
    is_flag=True, default=False,
    help='Include all entries, disregarding all other filter options and flags.')

ALL_STATES = OverridableOption(
    '-A', '--all-states',
    is_flag=True,
    help='Do not limit to items in running state.')

ALL_USERS = OverridableOption(
    '-A', '--all-users', 'all_users',
    is_flag=True, default=False,
    help='Include all entries regardless of the owner.')

GROUP_CLEAR = OverridableOption(
    '-c', '--clear',
    is_flag=True, default=False,
    help='Remove all the nodes from the group.')

RAW = OverridableOption(
    '-r', '--raw', 'raw',
    is_flag=True, default=False,
    help='Display only raw query results, without any headers or footers.')

HOSTNAME = OverridableOption(
    '-H', '--hostname',
    help='Hostname.')

TRANSPORT = OverridableOption(
    '-T', '--transport',
    type=types.PluginParamType(group='transports'), required=True,
    help='Transport type.')

SCHEDULER = OverridableOption(
    '-S', '--scheduler',
    type=types.PluginParamType(group='schedulers'), required=True,
    help='Scheduler type.')

USER = OverridableOption(
    '-u', '--user', 'user',
    type=types.UserParamType(),
    help='Email address of the user.')

PORT = OverridableOption(
    '-P', '--port', 'port',
    type=click.INT,
    help='Port number.')

FREQUENCY = OverridableOption(
    '-F', '--frequency', 'frequency',
    type=click.INT)

VERBOSE = OverridableOption(
    '-v', '--verbose',
    is_flag=True, default=False,
    help='Be more verbose in printing output.')

TIMEOUT = OverridableOption(
    '-t', '--timeout',
    type=click.FLOAT, default=5.0, show_default=True,
    help='Time in seconds to wait for a response before timing out.')

FORMULA_MODE = OverridableOption(
    '-f', '--formula-mode',
    type=click.Choice(['hill', 'hill_compact', 'reduce', 'group', 'count', 'count_compact']),
    default='hill',
    help='Mode for printing the chemical formula.')

TRAJECTORY_INDEX = OverridableOption(
    '-i', '--trajectory-index', 'trajectory_index',
    type=click.INT,
    default=None,
    help='Specific step of the Trajecotry to select.')

WITH_ELEMENTS = OverridableOption(
    '-e', '--with-elements', 'elements',
    type=click.STRING,
    cls=MultipleValueOption,
    default=None,
    help='Only select objects containing these elements.')

WITH_ELEMENTS_EXCLUSIVE = OverridableOption(
    '-E', '--with-elements-exclusive', 'elements_exclusive',
    type=click.STRING,
    cls=MultipleValueOption,
    default=None,
    help='Only select objects containing only these and no other elements.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Default values and lazy default get methods for command line options."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.cmdline.utils import echo
from aiida.common import exceptions
from aiida.manage.configuration import get_config


def get_default_profile(ctx, param, value):  # pylint: disable=unused-argument
    """Try to get the default profile.

    This should be used if the default profile should be returned lazily, at a point for example when the config
    is not created at import time. Otherwise, the preference should go to calling `get_config` to load the actual
    config and using `config.default_profile_name` to get the default profile name

    :raises click.UsageError: if the config could not be loaded or no default profile exists
    :return: the default profile
    """
    if value:
        return value

    try:
        config = get_config()
    except exceptions.ConfigurationError as exception:
        echo.echo_critical(str(exception))

    try:
        default_profile = config.get_profile(config.default_profile_name)
    except exceptions.ProfileConfigurationError:
        default_profile = None

    return default_profile



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""The `verdi data` command line interface."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.commands.cmd_plugin import verdi_plugin
from aiida.cmdline.utils import decorators
from aiida.cmdline.utils.pluginable import Pluginable


@verdi.group('data', entry_point_group='aiida.cmdline.data', cls=Pluginable)
def verdi_data():
    """Inspect, create and manage data nodes."""


@verdi_data.command('plugins')
@click.argument('entry_point', type=click.STRING, required=False)
@click.pass_context
@decorators.with_dbenv()
@decorators.deprecated_command("This command is deprecated. Use 'verdi plugin list' instead.")
def data_plugins(ctx, entry_point):
    """Print a list of registered data plugins or details of a specific data plugin."""
    ctx.invoke(verdi_plugin.get_command(ctx, 'list'), entry_point_group='aiida.data', entry_point=entry_point)
